import logging
import shlex
import subprocess
import time
import webbrowser
from pathlib import Path
from typing import Optional

import requests
import typer
from rich.console import Console

import runhouse as rh
import runhouse.rns.login

from runhouse import __version__, cluster, Cluster, configs

from runhouse.cli_utils import (
    add_clusters_to_output_table,
    check_ray_installation,
    create_output_table,
    get_local_or_remote_cluster,
    get_wrapped_server_start_cmd,
    is_command_available,
    LogsSince,
    print_cluster_config,
    print_status,
    StatusType,
)

from runhouse.constants import (
    BULLET_UNICODE,
    ITALIC_BOLD,
    MAX_CLUSTERS_DISPLAY,
    RAY_KILL_CMD,
    RAY_START_CMD,
    RESET_FORMAT,
    SERVER_LOGFILE,
    SERVER_STOP_CMD,
)
from runhouse.globals import rns_client

from runhouse.logger import get_logger

from runhouse.resources.hardware import (
    check_for_existing_ray_instance,
    get_all_sky_clusters,
    kill_actors,
)
from runhouse.resources.hardware.utils import ClusterStatus

from runhouse.servers.obj_store import ObjStoreError

SKY_LIVE_CLUSTERS_MSG = (
    "Live on-demand clusters created via Sky may exist that are not saved in Den. "
    "For more information, please run [bold italic]`sky status -r`[/bold italic]."
)

# create an explicit Typer application
app = typer.Typer(add_completion=False)

###############################
# Resources Command Groups
###############################

# creating a cluster app to enable subcommands of cluster (ex: runhouse cluster list).
# Register it with the main runhouse application
cluster_app = typer.Typer(help="Cluster related CLI commands.")

app.add_typer(cluster_app, name="cluster")

# creating a server app to enable subcommands of server (ex: runhouse server status).
# Register it with the main runhouse application
server_app = typer.Typer(help="Runhouse server related CLI commands.")
app.add_typer(server_app, name="server")

# For printing with typer
console = Console()

logger = get_logger(__name__)


###############################
# General Runhouse CLI commands
###############################
@app.command()
def login(
    token: Optional[str] = typer.Argument(None, help="Your Runhouse API token"),
    sync_secrets: Optional[bool] = typer.Option(
        False,
        "--sync-secrets",
        help="Whether to sync secrets. You will be prompted whether to upload local secrets or download saved secrets.",
    ),
    yes: Optional[bool] = typer.Option(
        False, "--yes", "-y", help="Sets any confirmations to 'yes' automatically."
    ),
):
    """Login to Runhouse. Validates token provided, with options to upload or download stored secrets or config between
    local environment and Runhouse / Vault.
    """
    valid_token: str = (
        runhouse.rns.login.login(
            token=token,
            download_config=True,
            upload_config=True,
            download_secrets=True,
            upload_secrets=True,
            from_cli=True,
        )
        if yes
        else runhouse.rns.login.login(
            token=token,
            interactive=True,
            ret_token=True,
            from_cli=True,
            sync_secrets=sync_secrets,
        )
    )

    if valid_token:
        webbrowser.open(f"{configs.get('dashboard_url')}/dashboard?token={valid_token}")
        raise typer.Exit()
    else:
        raise typer.Exit(code=1)


@app.command()
def logout():
    """Logout of Runhouse. Provides options to delete locally configured secrets and local Runhouse configs"""
    runhouse.rns.login.logout(interactive=True)
    raise typer.Exit()


###############################
# Cluster CLI commands
###############################
@cluster_app.command("ssh")
def cluster_ssh(
    cluster_name: str = typer.Argument(
        ...,
        help="Name of the cluster to SSH into.",
    ),
    node: Optional[str] = typer.Option(
        None,
        "-n",
        "--node",
        help="Specific cluster node to SSH into. If not specified will default to the head node.",
    ),
):
    """SSH into a remote cluster.

    Example:
        ``$ runhouse cluster ssh rh-basic-cpu``

    """
    try:
        c = cluster(name=cluster_name)
        if isinstance(c, rh.OnDemandCluster):
            if c.cluster_status == ClusterStatus.INITIALIZING:
                console.print(
                    f"[reset]{cluster_name} is being initialized. Please wait for it to finish, or run [reset][bold italic]`runhouse cluster up {cluster_name} -f`[/bold italic] to abort the initialization and relaunch."
                )
                raise typer.Exit(0)
            c.ssh(node=node)

        else:
            if node:
                raise ValueError(
                    "Node argument is only supported for on-demand clusters"
                )
            c.ssh()

    except ValueError:
        try:
            import sky

            state = sky.status(cluster_names=[cluster_name], refresh=False)
        except:
            state = []

        if len(state) == 0:
            console.print(
                "Cluster must either be saved to Den, shared with you, or be a local ondemand cluster "
                "that is currently up."
            )
            raise typer.Exit(1)

        subprocess.run(f"ssh {cluster_name}", shell=True)


@cluster_app.command("status")
def cluster_status(
    cluster_name: str = typer.Argument(
        None,
        help="Name of the cluster to check. If not specified will check the local cluster.",
    ),
    send_to_den: bool = typer.Option(
        default=False,
        help="Whether to update Den with the status.",
    ),
):
    """Load the status of the cluster.

    Example:
        ``$ runhouse cluster status rh-basic-cpu``
    """
    current_cluster = get_local_or_remote_cluster(cluster_name=cluster_name)

    try:
        cluster_status = current_cluster.status(send_to_den=send_to_den)
    except ValueError:
        console.print("Failed to load status for cluster.")
        return

    except ConnectionError:
        console.print(
            "\N{smiling face with horns} Runhouse Daemon is not running... \N{No Entry} \N{Runner}"
        )
        return

    print_status(cluster_status, current_cluster)


@cluster_app.command("list")
def cluster_list(
    show_all: bool = typer.Option(
        False,
        "-a",
        "--all",
        help=f"Get all clusters saved in Den. Up to {MAX_CLUSTERS_DISPLAY} most recently active clusters "
        f"will be displayed.",
    ),
    since: Optional[str] = typer.Option(
        None,
        "--since",
        help="Time duration to filter on. Minimum allowable filter is 1 minute. You may filter by seconds (s), "
        "minutes (m), hours (h) or days (s). Examples: 30s, 15m, 2h, 3d.",
    ),
    cluster_status: Optional[ClusterStatus] = typer.Option(
        None,
        "--status",
        help="Cluster status to filter on.",
    ),
    force: bool = typer.Option(
        False,
        "-f",
        "--force",
        help="Whether to force a status update for all relevant clusters, or load the latest values.",
    ),
):
    """
    Load Runhouse clusters

    Example:
        ``$ runhouse cluster list``

        ``$ runhouse cluster list --all``

        ``$ runhouse cluster list --status running``

        ``$ runhouse cluster list --since 15m``
    """

    # logged out case
    if not rh.configs.token:
        sky_cli_command_formatted = f"{ITALIC_BOLD}`sky status -r{RESET_FORMAT}`"  # will be printed bold and italic
        console.print(
            f"Listing clusters requires a Runhouse token. Please run `runhouse login` to get your token, "
            f"or run {sky_cli_command_formatted} to list locally stored on-demand clusters."
        )
        return

    clusters = Cluster.list(
        show_all=show_all, since=since, status=cluster_status, force=force
    )

    den_clusters = clusters.get("den_clusters", None)
    running_clusters = (
        [
            den_cluster
            for den_cluster in den_clusters
            if den_cluster.get("Status") == ClusterStatus.RUNNING
        ]
        if den_clusters
        else None
    )

    sky_clusters = clusters.get("sky_clusters", None)

    if not den_clusters:
        no_clusters_msg = f"No clusters found in Den for {rns_client.username}"
        if show_all or since or cluster_status:
            no_clusters_msg += " that match the provided filters"
        console.print(no_clusters_msg)
        if sky_clusters:
            console.print(SKY_LIVE_CLUSTERS_MSG)
        return

    filters_requested: bool = show_all or since or cluster_status

    clusters_to_print = den_clusters if filters_requested else running_clusters

    if show_all:
        # if user requesting all den cluster, limit print only to 50 clusters max.
        clusters_to_print = clusters_to_print[
            : (min(len(clusters_to_print), MAX_CLUSTERS_DISPLAY))
        ]

    # creating the clusters table
    table = create_output_table(
        total_clusters=len(den_clusters),
        running_clusters=len(running_clusters),
        displayed_clusters=len(clusters_to_print),
        filters_requested=filters_requested,
    )

    add_clusters_to_output_table(table=table, clusters=clusters_to_print)

    console.print(table)

    # print msg about un-saved live sky clusters.
    if sky_clusters:
        console.print(SKY_LIVE_CLUSTERS_MSG)


@cluster_app.command("keep-warm")
def cluster_keep_warm(
    cluster_name: str = typer.Argument(
        None,
        help="Name of cluster to keep warm. If not specified will check the local cluster.",
    ),
    mins: Optional[int] = typer.Option(
        -1,
        help="Amount of time (in min) to keep the cluster warm after inactivity. "
        "If set to -1, keeps cluster warm indefinitely.",
    ),
):
    """Keep the cluster warm for given number of minutes after inactivity.

    Example:
        ``$ runhouse cluster keep-warm rh-basic-cpu``

    """

    try:
        current_cluster = get_local_or_remote_cluster(cluster_name, exit_on_error=False)
        current_cluster.keep_warm(mins=mins)

    except ValueError:
        sky_live_clusters = get_all_sky_clusters()
        cluster_name = (
            cluster_name.split("/")[-1] if "/" in cluster_name else cluster_name
        )
        if cluster_name in sky_live_clusters:
            console.print(
                f"You can keep the cluster warm by running [italic bold]`sky autostop {cluster_name} -i {mins}`"
            )

    except Exception as e:
        console.print(f"Failed to keep the cluster warm: {e}")


@cluster_app.command("up")
def cluster_up(
    cluster_name: str = typer.Argument(
        None,
        help="The name of cluster to bring up.",
    ),
    force: bool = typer.Option(
        False,
        "-f",
        "--force",
        help="Whether to up the cluster regardless of its current initialization status",
    ),
):
    """Bring up the cluster if it is not up. No-op if cluster is already up.
    This only applies to on-demand clusters, and has no effect on self-managed clusters.

    Note: To launch the cluster via Den, set `launcher: den` in your local `~/.rh/config.yaml`.

    Example:
        ``$ runhouse cluster up rh-basic-cpu``
        ``$ runhouse cluster up rh-basic-cpu --force``

    """
    try:
        current_cluster = rh.cluster(name=cluster_name, dryrun=True)
        if current_cluster.is_up():
            console.print("Cluster is already up.")
            return
        current_cluster.up(force=force)
    except ValueError:
        console.print("Cluster not found in Den.")
    except Exception as e:
        console.print(f"Failed to bring up the cluster: {e}")


@cluster_app.command("down")
def cluster_down(
    cluster_name: str = typer.Argument(
        None,
        help="The name of cluster to terminate.",
    ),
    remove_configs: bool = typer.Option(
        False,
        "-rm",
        "--remove-configs",
        help="Whether to delete cluster config from Den (default: ``False``).",
    ),
    remove_all: bool = typer.Option(
        False,
        "-a",
        "--all",
        help="Whether to terminate all running clusters saved in Den (default: ``False``).",
    ),
    force_deletion: bool = typer.Option(
        False, "-y", "--yes", help="Skip confirmation prompt."
    ),
):
    """Terminate cluster if it is not down. No-op if cluster is already down.
    This only applies to on-demand clusters, and has no effect on self-managed clusters.

    Example:
        ``$ runhouse cluster down rh-basic-cpu``
    """
    if not force_deletion:
        if cluster_name:
            proceed = typer.prompt(
                f"Terminating {cluster_name}. Proceed? [Y/n] (Press Enter to terminate)",
                default="",
                show_default=False,
            )
        elif remove_all:
            proceed = typer.prompt(
                "Terminating all running clusters saved in Den. Proceed? [Y/n] (Press Enter to terminate)",
                default="",
                show_default=False,
            )
        else:
            console.print("Cannot determine which cluster to terminate, aborting.")
            raise typer.Exit(1)

        if proceed.lower() in ["n", "no"]:
            console.print("Aborted!")
            raise typer.Exit(0)

    if remove_all:
        running_den_clusters = Cluster.list(status=ClusterStatus.RUNNING).get(
            "den_clusters"
        )

        if not running_den_clusters:
            console.print("No running clusters saved in Den")
            raise typer.Exit(0)

        num_terminated_clusters: int = 0
        for running_cluster in running_den_clusters:
            try:
                current_cluster = rh.cluster(
                    name=f'{running_cluster.get("Name")}',
                    dryrun=True,
                )
                if isinstance(current_cluster, rh.OnDemandCluster):
                    current_cluster.teardown_and_delete() if remove_configs else current_cluster.teardown()
                    num_terminated_clusters += 1
                else:
                    console.print(
                        f"[reset][bold italic]{current_cluster.rns_address} [reset]is not an on-demand cluster and must be terminated manually."
                    )
            except ValueError:
                continue

        if num_terminated_clusters > 0:
            console.print(
                f"Successfully terminated [reset]{num_terminated_clusters} clusters."
            )

        raise typer.Exit(0)

    try:
        current_cluster: rh.Cluster = rh.cluster(name=cluster_name, dryrun=True)
        if isinstance(current_cluster, rh.OnDemandCluster):
            current_cluster.teardown_and_delete() if remove_configs else current_cluster.teardown()
        else:
            console.print(
                f"{current_cluster.rns_address} is not an on-demand cluster and must be terminated manually."
            )
    except ValueError:
        console.print("Cluster is not saved in Den, could not bring it down.")
        sky_live_clusters = get_all_sky_clusters()
        cluster_name = (
            cluster_name.split("/")[-1] if "/" in cluster_name else cluster_name
        )
        if cluster_name in sky_live_clusters:
            console.print(
                f"You can bring down the cluster by running [italic bold] `sky down {cluster_name}`"
            )
    except Exception as e:
        console.print(f"Failed to terminate the cluster: {e}")


@cluster_app.command("logs")
def cluster_logs(
    cluster_name: str = typer.Argument(
        None,
        help="Name of cluster to load the logs from. If not specified, loads the logs of the local cluster.",
    ),
    since: Optional[LogsSince] = typer.Option(
        LogsSince.five.value,
        "--since",
        help="Get the logs from the provided timeframe (in minutes).",
    ),
):
    """Load the logs of the Runhouse server running on a cluster.

    Example:
        ``$ runhouse cluster logs rh-basic-cpu``

        ``$ runhouse cluster logs rh-basic-cpu --since 60``

    """
    current_cluster = get_local_or_remote_cluster(cluster_name=cluster_name)

    cluster_uri = rns_client.resource_uri(current_cluster.rns_address)

    resp = requests.get(
        f"{rns_client.api_server_url}/resource/{cluster_uri}/logs/file?minutes={since}",
        headers=rns_client.request_headers(),
    )

    if resp.status_code == 404:
        console.print("Cluster logs are not available yet.")
        return
    if resp.status_code != 200:
        console.print("Failed to load cluster logs.")
        return

    logs_file_url = resp.json().get("data").get("logs_presigned_url")
    logs_file_content = requests.get(logs_file_url).content.decode("utf-8")

    stripped_lines = "\n".join(line.strip() for line in logs_file_content.splitlines())
    console.print("-" * len(current_cluster.rns_address))
    console.print(f"[reset][cyan]{current_cluster.rns_address}")
    console.print("-" * len(current_cluster.rns_address))
    console.print(f"[reset][cyan]{stripped_lines}")


###############################
# Server CLI commands
###############################
def _start_server(
    restart,
    restart_ray,
    screen,
    nohup,
    create_logfile=True,
    host=None,
    port=None,
    use_https=False,
    den_auth=False,
    ssl_keyfile=None,
    ssl_certfile=None,
    restart_proxy=False,
    use_caddy=False,
    domain=None,
    certs_address=None,
    api_server_url=None,
    conda_env=None,
    from_python=None,
):
    ############################################
    # Build CLI commands to start the server
    ############################################
    cmds = []
    if restart:
        cmds.append(SERVER_STOP_CMD)

    # We have to `ray start` not within screen/nohup
    existing_ray_instance = check_for_existing_ray_instance()
    if not existing_ray_instance or restart_ray:
        cmds.append(RAY_KILL_CMD)
        cmds.append(RAY_START_CMD)

    # Collect flags
    flags = []

    den_auth_flag = " --use-den-auth" if den_auth else ""
    if den_auth_flag:
        logger.info("Starting server with Den auth.")
        flags.append(den_auth_flag)

    restart_proxy_flag = " --restart-proxy" if restart_proxy else ""
    if restart_proxy_flag:
        logger.info("Reinstalling server configs.")
        flags.append(restart_proxy_flag)

    use_caddy_flag = " --use-caddy" if use_caddy else ""
    if use_caddy_flag:
        logger.info("Configuring Caddy on the cluster.")
        flags.append(use_caddy_flag)

    ssl_keyfile_flag = f" --ssl-keyfile {ssl_keyfile}" if ssl_keyfile else ""
    if ssl_keyfile_flag:
        logger.info(f"Using SSL keyfile in path: {ssl_keyfile}")
        flags.append(ssl_keyfile_flag)

    ssl_certfile_flag = f" --ssl-certfile {ssl_certfile}" if ssl_certfile else ""
    if ssl_certfile_flag:
        logger.info(f"Using SSL certfile in path: {ssl_certfile}")
        flags.append(ssl_certfile_flag)

    domain = f" --domain {domain}" if domain else ""
    if domain:
        logger.info(f"Using domain: {domain}")
        flags.append(domain)

    # Use HTTPS if explicitly specified or if SSL cert or keyfile path are provided
    https_flag = " --use-https" if use_https or (ssl_keyfile or ssl_certfile) else ""
    if https_flag:
        logger.info("Starting server with HTTPS.")
        flags.append(https_flag)

    host_flag = f" --host {host}" if host else ""
    if host_flag:
        logger.info(f"Using host: {host}.")
        flags.append(host_flag)

    port_flag = f" --port {port}" if port else ""
    if port_flag:
        logger.info(f"Using port: {port}.")
        flags.append(port_flag)

    address_flag = f" --certs-address {certs_address}" if certs_address else ""
    if address_flag:
        logger.info(f"Server public IP address: {certs_address}.")
        flags.append(address_flag)

    api_server_url_flag = (
        f" --api-server-url {api_server_url}" if api_server_url else ""
    )
    if api_server_url_flag:
        logger.info(f"Setting api_server url to {api_server_url}")
        flags.append(api_server_url_flag)

    conda_env_flag = f" --conda-env {conda_env}" if conda_env else ""
    if conda_env_flag:
        logger.info(f"Creating runtime env for conda env: {conda_env}")
        flags.append(conda_env_flag)

    flags.append(" --from-python" if from_python else "")

    # Check if screen or nohup are available
    screen = screen and is_command_available("screen")
    nohup = not screen and nohup and is_command_available("nohup")

    # Create logfile if we are using backgrounding
    if (screen or nohup) and create_logfile and not Path(SERVER_LOGFILE).exists():
        Path(SERVER_LOGFILE).parent.mkdir(parents=True, exist_ok=True)
        Path(SERVER_LOGFILE).touch()

    # Add flags to the server start command
    cmds.append(get_wrapped_server_start_cmd(flags, screen, nohup))
    logger.info(f"Starting API server using the following command: {cmds[-1]}.")

    try:
        # Open and read the lines of the server logfile so we only print the most recent lines after starting
        f = None
        if (screen or nohup) and Path(SERVER_LOGFILE).exists():
            f = open(SERVER_LOGFILE, "r")
            f.readlines()  # Discard these, they're from the previous times the server was started

        # We do these one by one so it's more obvious where the error is if there is one
        for i, cmd in enumerate(cmds):
            console.print(f"Executing `{cmd}`")
            if (
                i == len(cmds) - 1
            ):  # last cmd is not being parsed correctly when ran with shlex.split
                result = subprocess.run(cmd, shell=True, check=True)
            else:
                result = subprocess.run(shlex.split(cmd), text=True)

            if result.returncode != 0:
                # We don't want to raise an error if the server kill fails, as it may simply not be running
                if "pkill" in cmd:
                    continue

                # Retry ray start in case pkill process did not complete in time, up to 10s
                if cmd == RAY_START_CMD:
                    console.print("Retrying:")
                    attempt = 0
                    while result.returncode != 0 and attempt < 10:
                        attempt += 1
                        time.sleep(1)
                        result = subprocess.run(
                            shlex.split(cmd), text=True, capture_output=True
                        )
                        if result.stderr and "ConnectionError" not in result.stderr:
                            break
                    if result.returncode == 0:
                        continue

                console.print(f"Error while executing `{cmd}`")
                raise typer.Exit(1)

        server_started_str = "Uvicorn running on"
        # Read and print the server logs until the
        if screen or nohup:
            while not Path(SERVER_LOGFILE).exists():
                time.sleep(1)
            f = f or open(SERVER_LOGFILE, "r")
            start_time = time.time()
            # Wait for input for 60 seconds max (for Caddy to download and set up)
            while time.time() - start_time < 60:
                for line in f:
                    if server_started_str in line:
                        console.print(line)
                        f.close()
                        return
                    else:
                        console.print(line, end="")
                time.sleep(1)
            f.close()

    except FileNotFoundError:
        console.print(
            "python3 command was not found. Make sure you have python3 installed."
        )
        raise typer.Exit(1)


###############################
# Server CLI commands
###############################


@server_app.command("start")
def server_start(
    cluster_name: Optional[str] = typer.Argument(
        None,
        help="Specify a *saved* remote cluster to start the Runhouse server on that cluster. "
        "If not provided, the locally running server will be started.",
    ),
    restart_ray: bool = typer.Option(True, help="Restart the Ray runtime"),
    screen: bool = typer.Option(
        True,
        help="Start the server in a screen. Only relevant when restarting locally.",
    ),
    nohup: bool = typer.Option(
        True,
        help="Start the server in a nohup if screen is not available. Only relevant when restarting locally.",
    ),
    resync_rh: bool = typer.Option(
        False,
        help="Resync the Runhouse package. Only relevant when restarting remotely.",
    ),
    host: Optional[str] = typer.Option(
        None, help="Custom server host address. Default is `0.0.0.0`."
    ),
    port: Optional[str] = typer.Option(
        None, help="Port for server. If not specified will start on 32300."
    ),
    use_https: bool = typer.Option(
        False, help="Start an HTTPS server with TLS verification."
    ),
    use_den_auth: bool = typer.Option(
        False, help="Whether to authenticate requests with a Runhouse token."
    ),
    ssl_keyfile: Optional[str] = typer.Option(
        None, help="Path to custom SSL key file to use for enabling HTTPS."
    ),
    ssl_certfile: Optional[str] = typer.Option(
        None, help="Path to custom SSL cert file to use for enabling HTTPS."
    ),
    restart_proxy: bool = typer.Option(
        False, help="Whether to reinstall server configs on the cluster."
    ),
    use_caddy: bool = typer.Option(
        False,
        help="Whether to configure Caddy on the cluster as a reverse proxy.",
    ),
    domain: str = typer.Option(
        None,
        help="Server domain. Relevant if using Caddy to automate generating CA verified certs.",
    ),
    certs_address: Optional[str] = typer.Option(
        None,
        help="Public IP address of the server. Required for generating self-signed certs and enabling HTTPS.",
    ),
    api_server_url: str = typer.Option(
        default="https://api.run.house",
        help="URL of Runhouse Den",
    ),
    conda_env: str = typer.Option(
        None, help="Name of conda env where Runhouse server is started, if applicable."
    ),
    from_python: bool = typer.Option(
        False,
        help="Whether HTTP server started from inside a Python call rather than CLI.",
    ),
):
    """
    Start the HTTP server on the cluster.

    Example:
        ``$ runhouse server start``

        ``$ runhouse server start rh-cpu``
    """
    check_ray_installation()

    # If server is already up, ask the user to restart the server instead.
    if not cluster_name:
        server_status_cmd = "runhouse server status"
        result = subprocess.run(
            server_status_cmd, shell=True, capture_output=True, text=True
        )

        if result.returncode == 0 and "Server is up" in result.stdout:
            console.print(
                "Local Runhouse server is already running. To restart it, please "
                "run [italic bold]`runhouse server restart` [reset]with the relevant parameters."
            )
            raise typer.Exit(0)

    if cluster_name:
        c = get_local_or_remote_cluster(cluster_name=cluster_name)
        c.start_server(resync_rh=resync_rh, restart_ray=restart_ray)
        return

    _start_server(
        restart=False,
        restart_ray=restart_ray,
        screen=screen,
        nohup=nohup,
        create_logfile=True,
        host=host,
        port=port,
        use_https=use_https,
        den_auth=use_den_auth,
        ssl_keyfile=ssl_keyfile,
        ssl_certfile=ssl_certfile,
        restart_proxy=restart_proxy,
        use_caddy=use_caddy,
        domain=domain,
        certs_address=certs_address,
        api_server_url=api_server_url,
        conda_env=conda_env,
        from_python=from_python,
    )


@server_app.command("restart")
def server_restart(
    cluster_name: str = typer.Argument(
        None,
        help="Specify a *saved* remote cluster to restart the Runhouse server on that cluster. "
        "If not provided, the locally running server will be restarted.",
    ),
    restart_ray: bool = typer.Option(True, help="Restart the Ray runtime."),
    screen: bool = typer.Option(
        True,
        help="Start the server in a screen. Only relevant when restarting locally.",
    ),
    nohup: bool = typer.Option(
        True,
        help="Start the server in a nohup if screen is not available. Only relevant when restarting locally.",
    ),
    resync_rh: bool = typer.Option(
        False,
        help="Resync the Runhouse package. Only relevant when restarting remotely.",
    ),
    host: Optional[str] = typer.Option(
        None, help="Custom server host address. Default is `0.0.0.0`."
    ),
    port: Optional[str] = typer.Option(
        None, help="Port for server. If not specified will start on 32300."
    ),
    use_https: bool = typer.Option(
        False, help="Start an HTTPS server with TLS verification."
    ),
    use_den_auth: bool = typer.Option(
        False, help="Whether to authenticate requests with a Runhouse token."
    ),
    ssl_keyfile: Optional[str] = typer.Option(
        None, help="Path to custom SSL key file to use for enabling HTTPS."
    ),
    ssl_certfile: Optional[str] = typer.Option(
        None, help="Path to custom SSL cert file to use for enabling HTTPS."
    ),
    restart_proxy: bool = typer.Option(
        False, help="Whether to reinstall server configs on the cluster."
    ),
    use_caddy: bool = typer.Option(
        False,
        help="Whether to configure Caddy on the cluster as a reverse proxy.",
    ),
    domain: str = typer.Option(
        None,
        help="Server domain. Relevant if using Caddy to automate generating CA verified certs.",
    ),
    certs_address: Optional[str] = typer.Option(
        None,
        help="Public IP address of the server. Required for generating self-signed certs and enabling HTTPS.",
    ),
    api_server_url: str = typer.Option(
        default="https://api.run.house",
        help="URL of Runhouse Den",
    ),
    conda_env: str = typer.Option(
        None, help="Name of conda env where Runhouse server is started, if applicable."
    ),
    from_python: bool = typer.Option(
        False,
        help="Whether HTTP server started from inside a Python call rather than CLI.",
    ),
):
    """Restart the HTTP server on the cluster.

    Example:
        ``$ runhouse server restart``

        ``$ runhouse server restart rh-cpu``
    """
    check_ray_installation()

    if cluster_name:
        c = get_local_or_remote_cluster(cluster_name=cluster_name)
        c.restart_server(resync_rh=resync_rh, restart_ray=restart_ray)
        return

    _start_server(
        restart=True,
        restart_ray=restart_ray,
        screen=screen,
        nohup=nohup,
        create_logfile=True,
        host=host,
        port=port,
        use_https=use_https,
        den_auth=use_den_auth,
        ssl_keyfile=ssl_keyfile,
        ssl_certfile=ssl_certfile,
        restart_proxy=restart_proxy,
        use_caddy=use_caddy,
        domain=domain,
        certs_address=certs_address,
        api_server_url=api_server_url,
        conda_env=conda_env,
        from_python=from_python,
    )


@server_app.command("stop")
def server_stop(
    cluster_name: Optional[str] = typer.Argument(
        None,
        help="Specify a *saved* remote cluster to stop the Runhouse server on that cluster. "
        "If not provided, the locally running server will be stopped.",
    ),
    stop_ray: bool = typer.Option(False, help="Stop the Ray runtime."),
    cleanup_actors: bool = typer.Option(True, help="Kill all Ray actors."),
):
    """Stop the HTTP server on the cluster.

    Example:
        ``$ runhouse server stop``

        ``$ runhouse server stop rh-cpu``
    """
    check_ray_installation()
    logger.debug("Stopping the server")

    if cluster_name:
        current_cluster = get_local_or_remote_cluster(cluster_name=cluster_name)
        current_cluster.stop_server(stop_ray=stop_ray, cleanup_actors=cleanup_actors)
        return

    subprocess.run(SERVER_STOP_CMD, shell=True)

    if cleanup_actors:
        import ray

        ray.init(
            address="auto",
            ignore_reinit_error=True,
            logging_level=logging.ERROR,
            namespace="runhouse",
        )

        kill_actors(namespace="runhouse", gracefully=False)

    if stop_ray:
        logger.info("Stopping Ray.")
        subprocess.run(RAY_KILL_CMD, shell=True)

    console.print("Server stopped.")


@server_app.command("status")
def server_status(
    cluster_name: str = typer.Argument(
        None,
        help="Specify a *saved* remote cluster to check the status of the Runhouse server on that cluster. "
        "If not provided, the status of the locally running server will be checked.",
    ),
):
    """Check the HTTP server status on the cluster.

    Example:
        ``$ runhouse server status``

        ``$ runhouse server status rh-cpu``
    """
    check_ray_installation()
    logger.debug("Checking the server status.")
    current_cluster = get_local_or_remote_cluster(cluster_name=cluster_name)
    try:
        status = current_cluster.status()
        console.print(f"[reset]{BULLET_UNICODE} server pid: {status.get('server_pid')}")
        print_cluster_config(
            cluster_config=status.get("cluster_config"), status_type=StatusType.server
        )
        if rh.here == "file":
            console.print(
                "[reset]For more detailed information about the cluster, please use [italic bold]`runhouse cluster status <cluster_name>`[/italic bold]"
            )
    except (ObjStoreError, ConnectionError):
        console.print("Could not connect to Runhouse server. Is it up?")


@app.callback(invoke_without_command=True, help="Runhouse CLI")
def main(
    ctx: typer.Context,
    version: bool = typer.Option(
        None, "--version", "-v", help="Show the version and exit."
    ),
):
    """
    Runhouse CLI
    """
    if version:
        print(f"{__version__}")
    elif ctx.invoked_subcommand is None:
        subprocess.run("runhouse --help", shell=True)



import os

import runhouse as rh
import sky
from runhouse.rns.login import _login_download_secrets


def add_secrets_to_vault(headers):
    """Add some test secrets to Vault"""
    # Add real credentials for AWS and SKY to test sky status
    rh.provider_secret(
        name="/aws",  # add backslash / to name to force it to be vault secret
        provider="aws",
        values={
            "access_key": os.getenv("TEST_AWS_ACCESS_KEY"),
            "secret_key": os.getenv("TEST_AWS_SECRET_KEY"),
        },
    ).save(headers=headers)

    rh.provider_secret(
        name="/sky",
        provider="sky",
        values={
            "private_key": os.getenv("TEST_SKY_PRIVATE_KEY"),
            "public_key": os.getenv("TEST_SKY_PUBLIC_KEY"),
        },
    ).save(headers=headers)

    rh.provider_secret(
        name="/snowflake",
        provider="snowflake",
        values={"token": "ABCD1234"},
    ).save(headers=headers)


def test_login_flow_in_new_env():
    token = os.getenv("KITCHEN_TESTER_TOKEN")
    headers = {"Authorization": f"Bearer {token}"}

    add_secrets_to_vault(headers)

    secrets_in_vault = rh.Secret.vault_secrets(headers=headers)
    assert secrets_in_vault, "No secrets found in Vault"

    # Run login download secrets stored in Vault into the new env
    _login_download_secrets(headers=headers)

    # Once secrets are saved down to their local config, confirm we have sky enabled
    sky.check.check(quiet=True)
    clouds = sky.global_user_state.get_enabled_clouds()
    cloud_names = [str(c).lower() for c in clouds]
    assert "aws" in cloud_names

    for secret in secrets_in_vault.values():
        secret.delete(headers=headers)

    secrets_in_vault = rh.Secret.vault_secrets(headers=headers)
    assert not secrets_in_vault



import runhouse as rh


def sd_generate(
    prompt,
    num_images=1,
    steps=100,
    guidance_scale=7.5,
    model_id="stabilityai/stable-diffusion-2-base",
):
    import torch
    from diffusers import StableDiffusionPipeline

    pipe = StableDiffusionPipeline.from_pretrained(
        model_id, torch_dtype=torch.float16, revision="fp16"
    ).to("cuda")
    return pipe(
        [prompt] * num_images, num_inference_steps=steps, guidance_scale=guidance_scale
    ).images


def test_sd_generate(a10g_gpu_cluster):
    generate_gpu = rh.function(fn=sd_generate).to(
        a10g_gpu_cluster, reqs=["pytest", "diffusers", "torch", "transformers"]
    )

    images = generate_gpu(
        prompt="A hot dog made of matcha powder.", num_images=4, steps=50
    )
    assert images



import atexit

# Following design pattern for singleton variables from here:
# https://docs.python.org/3/faq/programming.html#how-do-i-share-global-variables-across-modules

from runhouse.rns.defaults import Defaults
from runhouse.rns.rns_client import RNSClient
from runhouse.servers.obj_store import ObjStore

# Configure the logger once
# TODO commenting out for now because this duplicates the logging config in the root logger
# logging.config.dictConfig(LOGGING_CONFIG)

configs = Defaults()

ssh_tunnel_cache = {}


def clean_up_ssh_connections():
    for _, v in ssh_tunnel_cache.items():
        v.terminate()


atexit.register(clean_up_ssh_connections)

rns_client = RNSClient(configs=configs)

# Note: this initalizes a dummy global object. The obj_store must
# be properly initialized by a servlet via initialize.
obj_store = ObjStore()



from runhouse.resources.secrets.provider_secrets.provider_secret import ProviderSecret


class DockerRegistrySecret(ProviderSecret):
    """
    .. note::
        To create a DockerRegistrySecret, please use the factory method
        :func:`provider_secret` with ``provider="docker"``.
    """

    _PROVIDER = "docker"
    _DEFAULT_ENV_VARS = {
        "username": "SKYPILOT_DOCKER_USERNAME",
        "password": "SKYPILOT_DOCKER_PASSWORD",
        "server": "SKYPILOT_DOCKER_SERVER",
    }

    @staticmethod
    def from_config(config: dict, dryrun: bool = False, _resolve_children: bool = True):
        return DockerRegistrySecret(**config, dryrun=dryrun)



import importlib
import json
import os
import shutil
from pathlib import Path
from typing import Any, Dict, Optional, Set, Union

import dotenv

import requests
from pydantic import BaseModel

from runhouse.logger import get_logger

from runhouse.rns.utils.api import (
    generate_uuid,
    load_resp_content,
    read_resp_data,
    remove_null_values_from_dict,
    ResourceAccess,
)

from runhouse.utils import locate_working_dir

logger = get_logger(__name__)


# This is a copy of the Pydantic model that we use to validate in Den
class ResourceStatusData(BaseModel):
    cluster_config: dict
    server_cpu_utilization: float
    server_gpu_utilization: Optional[float]
    server_memory_usage: Dict[str, Any]
    server_gpu_usage: Optional[Dict[str, Any]]
    env_servlet_processes: Dict[str, Dict[str, Any]]
    server_pid: int
    runhouse_version: str


class RNSClient:
    """Manage a particular resource with the runhouse database"""

    CORE_RNS_FIELDS = [
        "name",
        "resource_type",
        "visibility",
        "folder",
        "users",
    ]
    DEFAULT_FS = "file"

    def __init__(self, configs) -> None:
        self.run_stack = []
        self._configs = configs
        self._prev_folders = []

        self.rh_directory = str(Path(locate_working_dir()) / "rh")
        self.rh_builtins_directory = str(
            Path(importlib.util.find_spec("runhouse").origin).parent / "builtins"
        )

        # TODO allow users to register other base folders
        # Register all the directories in rh folder as rns base folders
        rns_base_folders = (
            [
                str(subdir)
                for subdir in Path(self.rh_directory).iterdir()
                if subdir.is_dir()
            ]
            if Path(self.rh_directory).exists()
            else []
        )
        rns_base_folders.append(
            str(Path(importlib.util.find_spec("runhouse").origin).parent / "builtins")
        )
        self._index_base_folders(rns_base_folders)
        self._current_folder = None

        self.session = requests.Session()

    @property
    def default_folder(self):
        return self._configs.default_folder

    @property
    def current_folder(self):
        return self._current_folder if self._current_folder else self.default_folder

    @current_folder.setter
    def current_folder(self, value):
        self._current_folder = value

    @property
    def token(self):
        return self._configs.token

    @property
    def username(self):
        return self._configs.username

    @property
    def launcher(self):
        return self._configs.launcher

    @property
    def api_server_url(self):
        url_as_env_var = os.getenv("API_SERVER_URL")
        if url_as_env_var:
            return url_as_env_var
        return self._configs.get("api_server_url", None)

    @property
    def autosave(self):
        return self._configs.get("autosave", False)

    @property
    def default_ssh_key(self):
        return self._configs.get("default_ssh_key", None)

    def _index_base_folders(self, lst):
        self.rns_base_folders = {}
        for folder in lst:
            config = self._load_config_from_local(path=folder)
            rns_path = str(Path(self.default_folder) / Path(folder).name)
            if config:
                rns_path = config.get("rns_address")
            self.rns_base_folders[rns_path] = folder

    @staticmethod
    def resource_uri(name):
        """URI used when querying the Den server"""
        from runhouse.rns.top_level_rns_fns import resolve_rns_path

        rns_address = resolve_rns_path(name)
        return RNSClient.format_rns_address(rns_address)

    @staticmethod
    def base_folder(resource_uri: str):
        """Top level folder associated with a resource."""
        if resource_uri is None:
            return None

        paths = resource_uri.split("/")
        return paths[1] if len(paths) > 1 else None

    @staticmethod
    def format_rns_address(rns_address: str):
        if rns_address.startswith("/"):
            rns_address = rns_address[1:]
        return rns_address.replace("/", ":")

    @staticmethod
    def local_to_remote_address(rns_address):
        return rns_address.replace("~", "@")

    def remote_to_local_address(self, rns_address):
        return rns_address.replace(self.default_folder, "~")

    def autosave_resources(self):
        return bool(self.autosave and self.token)

    def request_headers(
        self, resource_address: str = None, headers: dict = None
    ) -> Union[dict, None]:
        """Returns the authentication headers to use for requests made to Den or to a cluster.

        If the request is being made to Den, we simply construct the request headers with the user's existing
        Runhouse token.

        If the request is being made to (or from) a cluster, we generate a new unique token to prevent exposing the
        user's original Runhouse token on the cluster. This new token is based on the user's existing Den token and
        the Den address of the resource (or cluster API) they are attempting to access.

        For example, if userA tries to access a function on a cluster that was shared with them by userB, we generate a
        new token containing userA's Den token and top level directory associated with the
        resource (e.g. if the function has a rns address of: "/userB/some_func", the top level directory will be set
        to: "userB", indicating that the function's owner/namespace is userB).

        The updated token used in requests made by userA to the cluster would then look like:
        "hash(userA den token + resource top level directory) + resource top level directory + userA", where
        "hash" is a sha256 hash function.

        This method also ensures that each user <> resource relationship will yield the same token, allowing
        resource owners to more easily identify specific users consuming or accessing their resource.

        Args:
            resource_address (str, optional): Name of the top level directory or namespace that the resource
                is associated with. For example, if the Den address of a function is: "/userA/some_func", the resource
                address would be "userA". If not provided, we construct headers with the user's default Runhouse token.
            headers (dict, optional): Request headers to use for the request. If not provided, we use the default
                headers (i.e. use the default Runhouse token).

        Returns:
            request_headers (Union[dict, None]): The resulting request headers, or ``None`` if no headers
                are required for the request.

        """
        if headers == {}:
            # Support use case where we explicitly do not want to provide headers (e.g. requesting a cert)
            return None

        if headers is None:
            # Use the default headers (i.e. the user's original Den token)
            headers: dict = self._configs.request_headers

        if not headers:
            return None

        if "Authorization" not in headers:
            raise ValueError(
                "Invalid request headers provided, expected in format: {Authorization: Bearer <token>}"
            )

        if resource_address is None:
            # If a base dir is not specified assume we are not sending a request to a cluster
            return headers

        den_token = None
        auth_header = headers["Authorization"]
        token_parts = auth_header.split(" ")
        if len(token_parts) == 2 and token_parts[0] == "Bearer":
            den_token = token_parts[1]

        if den_token is None:
            raise ValueError(
                "Failed to extract token from request auth header. Expected in format: Bearer <token>"
            )

        hashed_token = self.cluster_token(resource_address)

        return {"Authorization": f"Bearer {hashed_token}"}

    def cluster_token(
        self, resource_address: str, username: str = None, den_token: str = None
    ):
        """Load the hashed token as generated in Den. Cache the token value in-memory for a given resource address.
        Optionally provide a username and den token instead of using the default values stored in local configs."""
        if resource_address and "/" in resource_address:
            # If provided as a full rns address, extract the top level directory
            resource_address = self.base_folder(resource_address)

        uri = f"{self.api_server_url}/auth/token/cluster"
        token_payload = {
            "resource_address": resource_address,
            "username": username or self._configs.username,
        }

        headers = (
            {"Authorization": f"Bearer {den_token}"}
            if den_token
            else self._configs.request_headers
        )
        resp = self.session.post(
            uri,
            data=json.dumps(token_payload),
            headers=headers,
        )
        if resp.status_code != 200:
            raise Exception(
                f"Received [{resp.status_code}] from Den POST '{uri}': Failed to load cluster token: {load_resp_content(resp)}"
            )

        resp_data = read_resp_data(resp)
        return resp_data.get("token")

    def validate_cluster_token(self, cluster_token: str, cluster_uri: str) -> bool:
        """Checks whether a particular cluster token is valid for the given cluster URI"""
        request_uri = self.resource_uri(cluster_uri)
        uri = f"{self.api_server_url}/auth/token/cluster/{request_uri}"
        resp = self.session.get(
            uri,
            headers={"Authorization": f"Bearer {cluster_token}"},
        )
        return resp.status_code == 200

    def resource_request_payload(self, payload) -> dict:
        payload = remove_null_values_from_dict(payload)
        data = {}
        for k, v in payload.copy().items():
            if k not in self.CORE_RNS_FIELDS:
                data[k] = v
                # if adding to data field remove as standalone field
                del payload[k]
        payload["data"] = data
        return payload

    def load_account_from_env(
        self, token_env_var="RH_TOKEN", usr_env_var="RH_USERNAME", dotenv_path=None
    ) -> Dict[str, str]:
        dotenv.load_dotenv(dotenv_path=dotenv_path)

        test_token = os.getenv(token_env_var)
        test_username = os.getenv(usr_env_var)
        if not (test_token and test_username):
            return None

        self._configs.token = test_token
        self._configs.username = test_username
        self._configs.default_folder = f"/{test_username}"

        # The client caches the folder that is used as the current folder variable, we clear this so it loads the new
        # folder when we switch accounts
        self._current_folder = None

        return {
            "token": self._configs.token,
            "username": self._configs.username,
            "default_folder": self._configs.default_folder,
        }

    def load_account_from_file(self) -> None:
        # Setting this to None causes it to be loaded from file upon next access
        self._configs.defaults_cache = None

        # Calling with .get explicitly loads from the config.yaml file
        self._configs.token = self._configs.get("token", None)
        self._configs.username = self._configs.get("username", None)
        self._configs.default_folder = self._configs.get("default_folder", None)

        # Same as above, for this to correctly load the account/folder from the new cache, it needs to be unset
        self._current_folder = None

    # Run Stack
    # ---------------------

    def start_run(self, run_obj: "Run"):
        self.run_stack.append(run_obj)

    def stop_run(self):
        return self.run_stack.pop()

    def current_run(self):
        if not self.run_stack:
            return None
        return self.run_stack[-1]

    def add_upstream_resource(self, name: str):
        """Add a resource's name to the current run's upstream artifact registry if it's being loaded"""
        current_run = self.current_run()
        if current_run:
            artifact_name = self.resolve_rns_path(name)
            current_run._register_upstream_artifact(artifact_name)

    def add_downstream_resource(self, name: str):
        """Add a resource's name to the current run's downstream artifact registry if it's being saved"""
        current_run = self.current_run()
        if current_run:
            artifact_name = self.resolve_rns_path(name)
            current_run._register_downstream_artifact(artifact_name)

    # ---------------------

    def grant_resource_access(
        self,
        *,
        rns_address: str,
        user_emails: list = None,
        access_level: ResourceAccess,
        notify_users: bool,
        headers: Optional[dict] = None,
    ):
        resource_uri = self.resource_uri(rns_address)
        headers = headers or self.request_headers()
        access_payload = {
            "users": user_emails,
            "access_level": access_level,
            "notify_users": notify_users,
        }
        uri = f"{self.api_server_url}/resource/{resource_uri}/users/access"
        resp = self.session.put(
            uri,
            data=json.dumps(access_payload),
            headers=headers,
        )
        if resp.status_code != 200:
            raise Exception(
                f"Received [{resp.status_code}] from Den PUT '{uri}': Failed to grant access to: {load_resp_content(resp)}"
            )

        resp_data: dict = read_resp_data(resp)
        added_users: dict = resp_data.get("added_users", {})
        new_users: dict = resp_data.get("new_users", {})
        valid_users: Set = resp_data.get("valid_users", set())

        return added_users, new_users, valid_users

    def load_config(
        self,
        name,
        load_from_den=True,
    ) -> dict:
        if not name:
            return {}

        from runhouse.resources.hardware.utils import _current_cluster

        if "/" not in name:
            name = f"{self.current_folder}/{name}"

        rns_address = self.resolve_rns_path(name)

        if rns_address == _current_cluster("name"):
            return _current_cluster("config")

        if rns_address[0] in ["~", "^"]:
            config = self._load_config_from_local(rns_address)
            if config:
                return config

        if load_from_den and rns_address.startswith("/"):
            request_headers = self.request_headers()
            if not request_headers:
                raise PermissionError(
                    "No Runhouse token provided. Try running `$ runhouse login` or visiting "
                    "https://run.house/login to retrieve a token. If calling via HTTP, please "
                    "provide a valid token in the Authorization header."
                )

            resource_uri = self.resource_uri(name)
            logger.debug(f"Attempting to load config for {rns_address} from Den.")
            uri = f"{self.api_server_url}/resource/{resource_uri}"
            resp = self.session.get(
                uri,
                headers=request_headers,
            )
            if resp.status_code != 200:
                logger.debug(
                    f"Received [{resp.status_code}] from Den GET '{uri}': No config found in Den: {load_resp_content(resp)}"
                )
                # No config found, so return empty config
                return {}

            config: dict = read_resp_data(resp)
            if config.get("data", None):
                config.update(config["data"])
                del config["data"]
            return config
        return {}

    def _load_config_from_local(self, rns_address=None, path=None) -> Optional[dict]:
        """Load config from local file"""
        # TODO should we handle remote filesystems, or throw an error if system != 'file'?
        if not path:
            path = self.locate(rns_address, resolve_path=False)
            if not path:
                return None
        config_path = Path(path) / "config.json"
        if not config_path.exists():
            return None

        logger.debug(f"Loading config from local file {config_path}")
        with open(config_path, "r") as f:
            try:
                config = json.load(f)
            except json.decoder.JSONDecodeError as e:
                logger.error(f"Error loading config from {config_path}: {e}")
                return None
        if rns_address:
            config["name"] = rns_address
        return config

    def get_rns_address_for_local_path(self, local_path):
        """Get Den address for local path"""
        try:
            rel_path = str(Path(local_path).relative_to(self.rh_directory))
            return "~/" + rel_path
        except ValueError:
            return None

    def save_config(self, resource, overwrite: bool = True):
        """Register the resource, saving it to local config folder and/or Den config store. Uses the resource's
        `self.config()` to generate the dict to save."""
        rns_address = resource.rns_address
        config = resource.config()

        if not overwrite and self.exists(rns_address):
            raise ValueError(
                f"Resource {rns_address} already exists and overwrite is False."
            )

        if rns_address is None:
            raise ValueError("A resource must have a name to be saved.")

        if rns_address[0] in ["~", "^"]:
            self._save_config_to_local(config, rns_address)

        if rns_address.startswith("/"):
            self._save_config_in_rns(config, rns_address)

    def _save_config_to_local(self, config: dict, rns_address: str):
        if not rns_address:
            raise ValueError("Cannot save resource without rns address or path.")
        resource_dir = Path(self.locate(rns_address, resolve_path=False))
        resource_dir.mkdir(parents=True, exist_ok=True)
        config_path = resource_dir / "config.json"
        with open(config_path, "w") as f:
            json.dump(config, f, indent=4)
        logger.info(f"Saving config for {rns_address} to: {config_path}")

    def _save_config_in_rns(self, config, resource_name):
        """Update or create resource config in database"""
        logger.info(f"Saving config for {resource_name} to Den")

        resource_uri = self.resource_uri(resource_name)
        put_uri = f"{self.api_server_url}/resource/{resource_uri}"

        payload = self.resource_request_payload(config)
        headers = self.request_headers()
        resp = self.session.put(put_uri, data=json.dumps(payload), headers=headers)
        if resp.status_code == 200:
            logger.debug(f"Config updated in Den for resource: {resource_uri}")
        elif resp.status_code == 422:  # No changes made to existing Resource
            logger.debug(
                f"Config for {resource_uri} has not changed, nothing to update"
            )
        elif resp.status_code == 404:  # Resource not found
            logger.debug(f"Saving new resource in Den for resource: {resource_uri}")
            # Resource does not yet exist, in which case we need to create from scratch
            post_uri = f"{self.api_server_url}/resource"
            resp = self.session.post(
                post_uri,
                data=json.dumps(payload),
                headers=headers,
            )
            if resp.status_code != 200:
                raise Exception(
                    f"Received [{resp.status_code}] from Den POST '{post_uri}': Failed to create new resource '{resource_uri}': {load_resp_content(resp)}"
                )
        else:
            raise Exception(
                f"Received [{resp.status_code}] from Den PUT '{put_uri}': Failed to save resource '{resource_uri}': {load_resp_content(resp)}"
            )

    def delete_configs(
        self,
        resource,
    ):
        rns_address = (
            resource.rns_address
            if hasattr(resource, "rns_address")
            else self.resolve_rns_path(resource)
        )
        if rns_address is None:
            logger.warning("No rns address exists for resource")
            return

        if rns_address[0] in ["~", "^"]:
            path = self.locate(rns_address, resolve_path=False)
            if path and Path(path).exists():
                shutil.rmtree(path)
            else:
                logger.info(
                    f"Cannot delete resource {rns_address}, could not find the local config."
                )

        if rns_address.startswith("/"):
            resource_uri = self.resource_uri(rns_address)
            uri = f"{self.api_server_url}/resource/{resource_uri}?delete_creds=false"
            resp = self.session.delete(uri, headers=self.request_headers())
            if resp.status_code != 200:
                logger.error(
                    f"Received [{resp.status_code}] from Den DELETE '{uri}': Failed to delete resource: {load_resp_content(resp)}"
                )
            else:
                logger.info(f"Successful Den Delete '{uri}'")

    def resolve_rns_data_resource_name(self, name: str):
        """If no name is explicitly provided for the data resource, we need to create one based on the relevant
        rns path. If name is None, return a hex uuid.
        For example: my_func -> my_username/my_func"
        """
        if name is None:
            return generate_uuid()
        rns_path = self.resolve_rns_path(name)
        if rns_path.startswith("~"):
            return rns_path[2:]
        # For the purposes of building the path to the underlying data resource we don't need the slash
        return rns_path.lstrip("/")

    #########################
    # Folder Operations
    #########################

    def resolve_rns_path(self, path: str):
        if path == ".":
            return self.current_folder
        if path.startswith("./"):
            return self.current_folder + "/" + path[len("./") :]
        # if path == '~':
        #     return '/rh'
        # if path.startswith('~/'):
        #     return '/rh/' + path[2:]
        # TODO break out paths for remote rns?
        if path == "@":
            return self.default_folder
        if path.startswith("@/"):
            return self.default_folder + "/" + path[len("@/") :]
        # if path == '^':
        #     return self.RH_BUILTINS_FOLDER
        # if path.startswith('^'):
        #     return self.RH_BUILTINS_FOLDER + '/' + path[1:]
        return path

    @staticmethod
    def split_rns_name_and_path(path: str):
        if "/" not in path:
            return path, None
        return Path(path).name, str(Path(path).parent)

    def exists(
        self,
        name_or_path,
        resource_type: str = None,
    ):
        config = self.load_config(name_or_path)
        if not config:
            return False
        if resource_type:
            return config.get("resource_type") == resource_type
        return True

    def locate(
        self,
        name,
        resolve_path=True,
    ):
        """Return the path for a resource."""
        # First check if name is in current folder

        if name == "/":
            return None

        if resolve_path:
            name = self.resolve_rns_path(name)

        if name.startswith("~"):
            return name.replace("~", self.rh_directory)

        if name.startswith("^"):
            return name.replace("^", self.rh_builtins_directory + "/")

        # TODO [DG] see if this breaks anything, also make it traverse the various rns folders to find the resource
        # if name.startswith('/'):
        #     if self.exists(name):
        #         return self.resource_uri(name)

        return None

    def set_folder(self, path: str, create=False):
        from runhouse.resources.folders import Folder, folder

        if isinstance(path, Folder):
            abs_path = path.rns_address
        else:
            abs_path = self.resolve_rns_path(path)
            if abs_path in ["~", "~/"]:
                create = False
            if create:
                folder(path=path, dryrun=False)

        self._prev_folders += [self.current_folder]
        self.current_folder = abs_path

    def unset_folder(self):
        """Sort of like `cd -`, but with a full stack of the previous folder's set. Resets the
        current_folder to the previous one on the stack, the current_folder right before the
        current one was set."""
        if len(self._prev_folders) == 0:
            # TODO should we be raising an error here?
            return
        self.current_folder = self._prev_folders.pop(-1)

    def contents(self, name_or_path, full_paths):
        from runhouse.resources.folders import folder

        folder_url = self.locate(name_or_path)
        return folder(name=name_or_path, path=folder_url).resources(
            full_paths=full_paths
        )



import importlib
from pathlib import Path

import pytest
import runhouse as rh
from runhouse.utils import capture_stdout

from tests.utils import get_pid_and_ray_node


class TestMultiNodeCluster:

    MAP_FIXTURES = {"resource": "cluster"}

    # setting release testing to run with on-demand cluster that not using docker image,
    # because the latter causing nightly release tests in CI to run for a very long time (does not happen locally).
    # TODO: [JL / SB]: check how we could make CI run with docker on-demand cluster

    UNIT = {"cluster": []}
    LOCAL = {"cluster": []}
    MINIMAL = {"cluster": []}
    RELEASE = {"cluster": ["multinode_k8s_cpu_cluster"]}
    MAXIMAL = {
        "cluster": [
            "multinode_k8s_cpu_cluster",
            "multinode_cpu_docker_conda_cluster",
            "multinode_gpu_cluster",
        ]
    }

    @pytest.mark.level("release")
    def test_rsync_and_ssh_onto_worker_node(self, cluster):
        worker_node = cluster.ips[-1]
        local_rh_package_path = Path(importlib.util.find_spec("runhouse").origin).parent

        local_rh_package_path = local_rh_package_path.parent
        dest_path = f"~/{local_rh_package_path.name}"

        # Rsync Runhouse package onto the worker node
        cluster.rsync(
            source=str(local_rh_package_path),
            dest=dest_path,
            up=True,
            node=worker_node,
            contents=True,
        )

        status_codes = cluster.run_bash([f"ls -l {dest_path}"], node=worker_node)
        assert status_codes[0][0] == 0

        assert "runhouse" in status_codes[0][1]

    @pytest.mark.level("release")
    def test_ray_started_on_worker_node_after_cluster_restart(self, cluster):
        head_node = cluster.head_ip

        status_codes = cluster.run_bash(["ray status"], node=head_node)
        assert status_codes[0][0] == 0

        status_output = status_codes[0][1]
        node_marker = "1 node_"
        num_nodes = status_output.count(node_marker)
        assert num_nodes == 2

    @pytest.mark.level("release")
    def test_send_envs_to_specific_worker_node(self, cluster):

        proc_0 = cluster.ensure_process_created("worker_0", compute={"node_idx": 0})
        proc_1 = cluster.ensure_process_created("worker_1", compute={"node_idx": 1})
        proc_2 = cluster.ensure_process_created("worker_2", compute={"node_idx": 1})

        with pytest.raises(ValueError):
            cluster.ensure_process_created(
                "worker_3", compute={"node_idx": len(cluster.ips)}
            )

        get_pid_0 = rh.function(get_pid_and_ray_node).to(
            name="get_pid_0", system=cluster, process=proc_0
        )
        get_pid_1 = rh.function(get_pid_and_ray_node).to(
            name="get_pid_1", system=cluster, process=proc_1
        )
        get_pid_2 = rh.function(get_pid_and_ray_node).to(
            name="get_pid_2", system=cluster, process=proc_2
        )

        with capture_stdout() as stdout_0:
            pid_0, node_id_0 = get_pid_0()
            assert str(pid_0) in str(stdout_0)
            assert str(node_id_0) in str(stdout_0)

        with capture_stdout() as stdout_1:
            pid_1, node_id_1 = get_pid_1()
            assert str(pid_1) in str(stdout_1)
            assert str(node_id_1) in str(stdout_1)

        with capture_stdout() as stdout_2:
            pid_2, node_id_2 = get_pid_2()
            assert str(pid_2) in str(stdout_2)
            assert str(node_id_2) in str(stdout_2)

        assert node_id_0 != node_id_1
        assert node_id_1 == node_id_2

    @pytest.mark.level("release")
    def test_specifying_resources(self, cluster):
        proc0 = cluster.ensure_process_created("worker_proc_0", compute={"CPU": 1.75})
        proc1 = cluster.ensure_process_created("worker_proc_1", compute={"CPU": 0.5})
        proc2 = cluster.ensure_process_created(
            "worker_proc_2", compute={"memory": 4 * 1024 * 1024 * 1024}
        )
        proc3 = cluster.ensure_process_created(
            "worker_proc_3", compute={"CPU": 0.1, "memory": 2 * 1024 * 1024 * 1024}
        )

        status = cluster.status()

        proc0_node = status["env_servlet_processes"][proc0]["node_ip"]
        proc1_node = status["env_servlet_processes"][proc1]["node_ip"]
        proc2_node = status["env_servlet_processes"][proc2]["node_ip"]
        proc3_node = status["env_servlet_processes"][proc3]["node_ip"]
        assert proc0_node in cluster.internal_ips
        assert proc1_node in cluster.internal_ips
        assert proc2_node in cluster.internal_ips
        assert proc3_node in cluster.internal_ips

        assert proc0_node != proc1_node  # Too much CPU
        assert proc2_node != proc3_node  # Too much memory

    @pytest.mark.level("release")
    def test_run_bash_on_node(self, cluster):
        # Specify via index
        result = cluster.run_bash("echo 'Hello World!'", node=0)
        assert result[0] == 0
        assert result[1] == "Hello World!\n"

        # Specify via IP
        result = cluster.run_bash("echo 'Hello World!'", node=cluster.ips[1])
        assert result[0] == 0
        assert result[1] == "Hello World!\n"

        # Run in process
        process = cluster.ensure_process_created("worker_env_0")
        result = cluster.run_bash("echo 'Hello World!'", process=process)
        assert result[0] == 0
        assert result[1] == "Hello World!\n"

    @pytest.mark.level("release")
    def test_multinode_secrets_to(self, cluster):
        custom_provider_secret = rh.provider_secret(
            provider="custom", values={"secret": "value"}
        )
        custom_provider_secret.to(cluster, path="~/.custom/secret.json")
        for node in cluster.ips:
            result = cluster.run_bash("ls ~/.custom", node=node)
            assert "secret.json" in result[1]

    @pytest.mark.level("release")
    def test_head_to_all_rsync(self, cluster):
        path = "/tmp/runhouse"
        cluster.run_bash(
            [f"mkdir -p {path}", f"echo 'hello there' > {path}/hello.txt"],
            node=cluster.head_ip,
            stream_logs=True,
        )

        cluster.rsync(
            source=path,
            dest=path,
            src_node=cluster.head_ip,
            node="all",
            contents=True,
            parallel=True,
        )

        for node in cluster.ips:
            status_codes = cluster.run_bash([f"ls -l {path}"], node=node)
            assert status_codes[0][0] == 0
            assert "hello.txt" in status_codes[0][1]
            status_codes = cluster.run_bash([f"cat {path}/hello.txt"], node=node)
            assert status_codes[0][0] == 0
            assert "hello there" in status_codes[0][1]

        # Test in the opposite direction, from worker to head, and parallel=False
        path = "/tmp/runhouse_2"
        cluster.run_bash(
            [f"mkdir -p {path}", f"echo 'hello again' > {path}/hello_again.txt"],
            node=cluster.ips[1],
            stream_logs=True,
        )

        cluster.rsync(
            source=path,
            dest=path,
            src_node=cluster.ips[1],
            node=cluster.head_ip,
            contents=True,
            parallel=False,
        )

        status_codes = cluster.run_bash([f"ls -l {path}"], node=cluster.head_ip)
        assert status_codes[0][0] == 0
        assert "hello_again.txt" in status_codes[0][1]
        status_codes = cluster.run_bash(
            [f"cat {path}/hello_again.txt"], node=cluster.head_ip
        )
        assert status_codes[0][0] == 0
        assert "hello again" in status_codes[0][1]



import logging
import sys

from runhouse.constants import DEFAULT_PROCESS_NAME

from runhouse.globals import configs, obj_store, rns_client
from runhouse.logger import get_logger
from runhouse.servers.obj_store import ClusterServletSetupOption

logger = get_logger(__name__)

logging.getLogger("numexpr").setLevel(logging.WARNING)


def resolve_rns_path(path: str):
    return rns_client.resolve_rns_path(path)


def exists(name, resource_type: str = None):
    return rns_client.exists(name, resource_type=resource_type)


def locate(name_or_path, resolve_path: bool = True):
    return rns_client.locate(
        name_or_path,
        resolve_path=resolve_path,
    )


def load(name: str, instantiate: bool = True, dryrun: bool = False):
    config = rns_client.load_config(name=name)
    if not instantiate:
        return config
    resource_class = getattr(
        sys.modules["runhouse"], config["resource_type"].capitalize(), None
    )
    if not resource_class:
        raise TypeError(
            f"Could not find module associated with {config['resource_type']}"
        )

    try:
        loaded = resource_class.from_config(config=config, dryrun=dryrun)
        rns_client.add_upstream_resource(name)
        return loaded
    except:
        raise ValueError(
            f"Could not find constructor for type {config['resource_type']}"
        )


async def get_local_cluster_object():
    # By default, obj_store.initialize does not initialize Ray, and instead
    # attempts to connect to an existing cluster.

    # In case we are calling `rh.here` within the same Python process
    # as an initialized object store, keep the same name.
    # If it was not set, let's proxy requests to `base` since we're likely on the cluster
    # and want to easily read and write from the object store that the Server is using.
    try:
        await obj_store.ainitialize(
            servlet_name=obj_store.servlet_name or DEFAULT_PROCESS_NAME,
            setup_cluster_servlet=ClusterServletSetupOption.GET_OR_FAIL,
        )
    except ConnectionError:
        return "file"

    # When HTTPServer is initialized, the cluster_config is set
    # within the global state.
    config = await obj_store.aget_cluster_config()
    if config.get("resource_subtype") is not None:
        from runhouse.resources.hardware.utils import _get_cluster_from

        system = _get_cluster_from(config, dryrun=True)
        return system

    return "file"


def save(resource, name: str = None, overwrite: bool = True, folder: str = None):
    """Register the resource, saving it to local working_dir config and/or Den config store. Uses the resource's
    `self.config()` to generate the dict to save."""

    # TODO handle self.access == 'read' instead of this weird overwrite argument
    if name:
        if "/" in name[1:]:
            (
                resource._name,
                resource._rns_folder,
            ) = split_rns_name_and_path(resolve_rns_path(name))
        else:
            resource._name = name
    if not resource.rns_address:
        resource._rns_folder = folder or rns_client.current_folder

    rns_client.save_config(resource=resource, overwrite=overwrite)


def set_folder(path: str, create=False):
    rns_client.set_folder(path=path, create=create)


def unset_folder():
    """Sort of like `cd -`, but with a full stack of the previous folder's set. Resets the
    current_folder to the previous one on the stack, the current_folder right before the
    current one was set."""
    rns_client.unset_folder()


def current_folder():
    return rns_client.current_folder


def split_rns_name_and_path(path: str):
    return rns_client.split_rns_name_and_path(path)


def resources(path: str = None, full_paths=False):
    path = path or current_folder()
    return rns_client.contents(name_or_path=path, full_paths=full_paths)


def ipython():
    try:
        import IPython

        IPython.embed()
    except ImportError:
        import code

        code.interact(local=locals())


def delete(resource_or_name: str):
    """Delete the resource from the Den or local config store."""
    rns_client.delete_configs(resource=resource_or_name)


class TokenContextManager:
    def __enter__(self):
        configs._use_caller_token = True

    def __exit__(self, exc_type, exc_val, exc_tb):
        configs._use_caller_token = False


def as_caller():
    return TokenContextManager()



from runhouse.resources.secrets.provider_secrets.api_key_secret import ApiKeySecret


class PineconeSecret(ApiKeySecret):
    """
    .. note::
            To create an PineconeSecret, please use the factory method :func:`provider_secret`
            with ``provider="pinecone"``.
    """

    _PROVIDER = "pinecone"
    _DEFAULT_ENV_VARS = {"api_key": "PINECONE_API_KEY"}

    @staticmethod
    def from_config(config: dict, dryrun: bool = False, _resolve_children: bool = True):
        return PineconeSecret(**config, dryrun=dryrun)



import inspect
import json
import os
import site
import time

from importlib import reload as importlib_reload

import numpy as np
import pandas as pd
import pytest

import runhouse as rh
from runhouse import Package
from runhouse.logger import get_logger
from runhouse.utils import capture_stdout

from tests.constants import TEST_ORG

logger = get_logger(__name__)

""" Tests for runhouse.Module. Structure:
    - Test call_module_method rpc, with various envs
    - Test creating module from class
    - Test creating module as rh.Module subclass
    - Test calling Module methods async
"""


def resolve_test_helper(obj):
    return obj


class SlowNumpyArray:
    def __init__(self, size=5):
        self.size = size
        self.arr = np.zeros(self.size)
        self._hidden_1 = "hidden"

    def slow_iter(self):
        self._hidden_2 = "hidden"
        if not self._hidden_1 and self._hidden_2:
            raise ValueError("Hidden attributes not set")
        for i in range(self.size):
            time.sleep(1)
            print(f"Hello from the cluster stdout! {i}")
            logger.info(f"Hello from the cluster logs! {i}")
            self.arr[i] = i
            yield f"Hello from the cluster! {self.arr}"

    def print_and_log(self, i):
        print(f"Hello from the cluster stdout! {i}")
        logger.info(f"Hello from the cluster logs! {i}")
        self.arr[i] = i
        return f"Hello from the cluster! {self.arr}"

    @classmethod
    def local_home(cls, local=True):
        return os.path.expanduser("~")

    @classmethod
    def home(cls):
        return os.path.expanduser("~")

    @classmethod
    def cpu_count(cls):
        return os.cpu_count()

    def size_minus_cpus(self):
        return self.size - self.cpu_count()

    @classmethod
    def factory_constructor(cls, size=5):
        return cls(size=size)


class SlowPandas(rh.Module):
    def __init__(self, size=5):
        super().__init__()
        self.size = size
        self.df = pd.DataFrame(np.zeros((self.size, self.size)))
        self._hidden_1 = "hidden"

    def slow_iter(self):
        for i in range(self.size):
            time.sleep(1)
            print(f"Hello from the cluster stdout! {i}")
            logger.info(f"Hello from the cluster logs! {i}")
            self.df[i] = i
            yield f"Hello from the cluster! {self.df.loc[[i]]}"

    async def slow_iter_async(self):
        for i in range(self.size):
            time.sleep(1)
            print(f"Hello from the cluster stdout! {i}")
            logger.info(f"Hello from the cluster logs! {i}")
            self.df[i] = i
            yield f"Hello from the cluster! {self.df.loc[[i]]}"

    @classmethod
    def local_home(cls, local=True):
        return os.path.expanduser("~")

    @classmethod
    def home(cls):
        return os.path.expanduser("~")

    @classmethod
    def cpu_count(cls):
        return os.cpu_count()

    def size_minus_cpus(self):
        return self.size - self.cpu_count()

    async def cpu_count_async(self):
        return os.cpu_count()


class ModuleConstructingOtherModule:
    def construct_and_get_process(self):
        remote_calc = rh.module(Calculator)()
        return remote_calc.process


class Calculator:
    importer = "Calculators Inc"

    def __init__(self, owner=None):
        self.model = "Casio"
        self._release_year = 2023
        self.owner = owner

    async def asummer(self, a: int, b: int):
        return a + b

    def summer(self, a: int, b: int):
        return a + b

    def sub(self, a: int, b: int):
        return a - b

    def divider(self, a: int, b: int):
        if b == 0:
            raise ZeroDivisionError
        return round(a / b)

    def mult(self, a: int, b: int):
        return a * b


class LotsOfAsync:
    async def asum(self, a: int, b: int):
        return a + b

    def returns_coroutine(self, a: int, b: int):
        return self.asum(a, b)

    async def async_returns_coroutine(self, a: int, b: int):
        return self.asum(a, b)


class ConstructorModule:
    def construct_module_on_cluster(self):
        calc_module = rh.module(Calculator)()
        assert calc_module.summer(2, 3) == 5


def load_and_use_readonly_module(mod_name, cpu_count, size=3):
    remote_df = rh.module(name=mod_name)
    # Check that module is readonly and cluster is not set
    assert isinstance(remote_df.system, str)
    assert remote_df.access_level == "read"

    assert remote_df.remote.size == size
    assert len(remote_df.remote.df) == size
    assert remote_df.remote._hidden_1 == "hidden"

    results = []
    # Capture stdout to check that it's working
    out = ""
    with capture_stdout() as stdout:
        for i, val in enumerate(remote_df.slow_iter()):
            assert val
            print(val)
            results += [val]
            out = out + str(stdout)
    assert len(results) == 3

    # Check that stdout was captured. Skip the last result because sometimes we
    # don't catch it and it makes the test flaky.
    for i in range(size - 1):
        assert f"Hello from the cluster stdout! {i}" in out
        assert f"Hello from the cluster logs! {i}" in out

    print(remote_df.cpu_count())
    assert remote_df.cpu_count() == os.cpu_count()
    print(remote_df.cpu_count(local=False))
    assert remote_df.cpu_count(local=False) == cpu_count

    # Test setting and getting properties
    df = remote_df.remote.df
    assert isinstance(df, pd.DataFrame)
    assert df.shape == (3, 3)
    assert df.loc[0, 0] == 0
    assert df.loc[2, 2] == 2

    remote_df.size = 20
    assert remote_df.remote.size == 20
    remote_df.size = size  # reset to original value for second test


async def load_and_use_calculator_module(mod_name):
    remote_calc = rh.module(name=mod_name)

    assert remote_calc.remote.model == "Casio"
    assert remote_calc.remote._release_year == 2023

    assert remote_calc.summer(2, 3) == 5
    assert await remote_calc.asummer(2, 3) == 5
    assert remote_calc.mult(3, 7) == 21

    assert remote_calc.sub(33, 5) == 28

    assert remote_calc.divider(16, 2) == 8


def construct_and_use_calculator_module(mod_name):
    remote_calc_class = rh.module(name=mod_name)
    remote_calc = remote_calc_class()

    assert remote_calc.remote.model == "Casio"
    assert remote_calc.remote._release_year == 2023

    assert remote_calc.summer(2, 3) == 5
    assert remote_calc.mult(3, 7) == 21
    assert remote_calc.sub(33, 5) == 28
    assert remote_calc.divider(16, 2) == 8

    return remote_calc


def nested_call_logs_stream_helper(slow_numpy_array):
    vals = []
    print("Starting nested call")
    time.sleep(1)
    for i in range(3):
        time.sleep(1)
        val = slow_numpy_array.print_and_log(i)
        vals.append(val)
    return vals


@pytest.mark.moduletest
class TestModule:

    MAP_FIXTURES = {"resource": "cluster"}

    UNIT = {"cluster": ["named_cluster"]}
    LOCAL = {
        "cluster": [
            "docker_cluster_pk_ssh_no_auth",  # Represents private dev use case
            "docker_cluster_pk_ssh_den_auth",  # Helps isolate Auth issues
            "docker_cluster_pk_http_exposed",  # Represents within VPC use case
            "docker_cluster_pwd_ssh_no_auth",
        ],
    }
    MINIMAL = {
        "cluster": [
            "static_cpu_pwd_cluster",
        ]
    }
    RELEASE = {
        "cluster": [
            "static_cpu_pwd_cluster_den_launcher",  # tests modules on a den-launched cpu cluster
            "static_cpu_pwd_cluster",  # tests modules on a local (sky) -launched cpu cluster
        ]
    }
    MAXIMAL = {
        "cluster": [
            "docker_cluster_pk_ssh_no_auth",
            "docker_cluster_pk_ssh_den_auth",
            "docker_cluster_pwd_ssh_no_auth",
            "static_cpu_pwd_cluster",
            "multinode_cpu_docker_conda_cluster",
            "static_gpu_pwd_cluster_den_launcher",  # tests modules on a den-launched cpu cluster
            "static_cpu_pwd_cluster_den_launcher",  # tests modules on a den-launched cpu cluster
        ],
    }

    # --------- integration tests ---------
    @pytest.mark.level("local")
    def test_call_module_method(self, cluster):
        cluster.put("numpy_pkg", Package.from_string("numpy"))

        # Test for property
        res = cluster.call("numpy_pkg", "config", stream_logs=True)
        numpy_config = Package.from_string("numpy").config()
        assert res
        assert isinstance(res, dict)
        assert res == numpy_config

        # Test iterator
        cluster.put("config_dict", list(numpy_config.keys()))
        res = cluster.call("config_dict", "__iter__", stream_logs=True)
        # Checks that all the keys in numpy_config were returned
        inspect.isgenerator(res)
        for key in res:
            assert key
            numpy_config.pop(key)
        assert not numpy_config

    @pytest.mark.level("local")
    def test_module_from_factory(self, cluster):
        size = 3
        RemoteClass = rh.module(SlowNumpyArray).to(cluster)
        remote_instance = RemoteClass(size=size, name="remote_instance1")

        # Test that naming works properly, and "class" module was unaffacted
        assert remote_instance.name == "remote_instance1"
        assert RemoteClass.name == "SlowNumpyArray"

        # Test that module was initialized correctly on the cluster
        assert remote_instance.system == cluster
        assert remote_instance.remote.size == size
        assert all(remote_instance.remote.arr == np.zeros(size))
        assert remote_instance.remote._hidden_1 == "hidden"

        results = []
        out = ""
        with capture_stdout() as stdout:
            for val in remote_instance.slow_iter():
                assert val
                print(val)
                results += [val]
                out = out + str(stdout)
        assert len(results) == 3

        # Check that stdout was captured. Skip the last result because sometimes we
        # don't catch it and it makes the test flaky.
        for i in range(size - 1):
            assert f"Hello from the cluster stdout! {i}" in out
            assert f"Hello from the cluster logs! {i}" in out

        local_home = os.path.expanduser("~")
        if cluster.on_this_cluster():
            remote_home = local_home
        else:
            remote_home = cluster.run_bash(["echo $HOME"])[0][1].strip()

        # Test classmethod on remote class
        assert RemoteClass.local_home() == local_home
        assert RemoteClass.home() == remote_home

        # Test classmethod on remote instance
        assert remote_instance.local_home() == local_home
        assert remote_instance.home() == remote_home

        # Test instance method
        assert remote_instance.size_minus_cpus() == size - remote_instance.cpu_count()

        # Test remote getter
        arr = remote_instance.remote.arr
        assert isinstance(arr, np.ndarray)
        assert arr.shape == (size,)
        assert arr[0] == 0
        assert arr[2] == 2

        # Test remote setter
        remote_instance.remote.size = 20
        assert remote_instance.remote.size == 20

        # Test creating a second instance of the same class
        remote_instance2 = RemoteClass(size=30, name="remote_instance2")
        assert remote_instance2.system == cluster
        assert remote_instance2.remote.size == 30

        # TODO fix factory method support
        # Test creating a third instance with the factory method
        remote_instance3 = RemoteClass.factory_constructor.remote(
            size=40, run_name="remote_instance3"
        )

        remote_instance_config = remote_instance3.system.config()
        cluster_config = cluster.config()

        assert remote_instance_config == cluster_config
        assert remote_instance3.remote.size == 40
        assert remote_instance3.home() == remote_home

        # Make sure first array and class are unaffected by this change
        assert remote_instance.remote.size == 20
        assert RemoteClass.home() == remote_home

        # Test that resolve() replaces the object properly on the cluster
        # TODO deprecate resolve and test that stub module is deserialized properly on the cluster as the true object
        #  in the obj_store
        # helper = rh.function(resolve_test_helper).to(cluster)
        # resolved_obj = helper(remote_instance.resolve())
        # assert resolved_obj.__class__.__name__ == "SlowNumpyArray"
        # assert not hasattr(resolved_obj, "config")
        # assert resolved_obj.size == 20
        # assert list(resolved_obj.arr) == [0, 1, 2]

    @pytest.mark.level("local")
    def test_module_from_subclass(self, cluster):
        size = 3
        remote_instance = SlowPandas(size=size).to(cluster)
        assert remote_instance.system == cluster

        # Test that module was initialized correctly on the cluster
        assert remote_instance.remote.size == size
        assert len(remote_instance.remote.df) == size
        assert remote_instance.remote._hidden_1 == "hidden"

        results = []
        # Capture stdout to check that it's working
        out = ""
        with capture_stdout() as stdout:
            for i, val in enumerate(remote_instance.slow_iter()):
                assert val
                print(val)
                results += [val]
                out = out + str(stdout)
        assert len(results) == 3

        # Check that stdout was captured. Skip the last result because sometimes we
        # don't catch it and it makes the test flaky.
        for i in range(remote_instance.size - 1):
            assert f"Hello from the cluster stdout! {i}" in out
            assert f"Hello from the cluster logs! {i}" in out

        local_home = os.path.expanduser("~")
        if cluster.on_this_cluster():
            remote_home = local_home
        else:
            remote_home = cluster.run_bash(["echo $HOME"])[0][1].strip()

        # Test classmethod on remote instance
        assert remote_instance.local_home() == local_home
        assert remote_instance.home() == remote_home

        # Test instance method
        assert remote_instance.size_minus_cpus() == size - remote_instance.cpu_count()

        # Test setting and getting properties
        df = remote_instance.remote.df
        assert isinstance(df, pd.DataFrame)
        assert df.shape == (3, 3)
        assert df.loc[0, 0] == 0
        assert df.loc[2, 2] == 2

        remote_instance.remote.size = 20
        assert remote_instance.remote.size == 20

        del remote_instance

        # Test get_or_to
        remote_instance = SlowPandas(size=3).get_or_to(cluster, name="SlowPandas")
        assert remote_instance.system.config() == cluster.config()
        # Check that size is unchanged from when we set it to 20 above
        assert remote_instance.remote.size == 20

        # Test that resolve() has no effect
        # TODO deprecate resolve and test that stub module is deserialized properly on the cluster as the true object
        #  in the obj_store
        # helper = rh.function(resolve_test_helper).to(cluster)
        # resolved_obj = helper(remote_instance.resolve())
        # assert resolved_obj.__class__.__name__ == "SlowPandas"
        # assert resolved_obj.size == 20  # resolved_obj.remote.size causing an error
        # assert resolved_obj.config() == remote_instance.config()

    @pytest.mark.level("local")
    @pytest.mark.asyncio
    async def test_module_from_subclass_async(self, cluster):
        remote_df = SlowPandas(size=3).to(cluster)
        assert remote_df.system == cluster
        results = []
        # Capture stdout to check that it's working
        out = ""
        with capture_stdout() as stdout:
            async for val in remote_df.slow_iter_async():
                assert val
                print(val)
                results += [val]
                out = out + str(stdout)
        assert len(results) == 3

        # Check that stdout was captured. Skip the last result because sometimes we
        # don't catch it and it makes the test flaky.
        # for i in range(remote_df.size - 1):
        #     assert f"Hello from the cluster stdout! {i}" in out
        #     assert f"Hello from the cluster logs! {i}" in out

        if cluster.on_this_cluster():
            cpu_count = os.cpu_count()
        else:
            cpu_count = int(
                cluster.run_python(["import os; print(os.cpu_count())"])[0][1]
            )
        assert await remote_df.cpu_count_async() == cpu_count

        # Properties
        df = await remote_df.fetch_async("df")
        assert isinstance(df, pd.DataFrame)
        assert df.shape == (3, 3)
        assert df.loc[0, 0] == 0
        assert df.loc[2, 2] == 2

        await remote_df.set_async("size", 20)
        assert remote_df.remote.size == 20

    @pytest.mark.skip("Not working yet")
    @pytest.mark.level("local")
    def test_hf_autotokenizer(self, cluster):
        from transformers import AutoTokenizer

        AutoTokenizer.from_pretrained("bert-base-uncased")
        cluster.install_package("transformers")
        RemoteAutoTokenizer = rh.module(AutoTokenizer).to(cluster)
        tok = RemoteAutoTokenizer.from_pretrained.remote(
            "bert-base-uncased", run_name="bert-tok"
        )
        # assert tok.remote.pad_token == "<pad>"
        prompt = "Tell me about unified development interfaces into compute and data infrastructure."
        assert tok(prompt, return_tensors="pt").shape == (1, 18)

    # --------- Unittests ---------

    @pytest.mark.usefixtures("cluster")
    @pytest.mark.level("local")
    def test_create_and_rename(self, cluster):
        RemoteCalc = rh.module(cls=Calculator).to(cluster)
        remote_calc = RemoteCalc(owner="Runhouse", name="Runhouse_calc")
        assert isinstance(RemoteCalc, rh.Module)
        assert RemoteCalc.name == "Calculator"
        assert remote_calc.name == "Runhouse_calc"
        assert remote_calc.summer(2, 3) == 5
        assert remote_calc.mult(3, 7) == 21

        # test that after rename, module stays the same, apart from the name.
        # Moreover, tests that the rename did not affect the existing remote instances, as well as creating new ones.
        RemoteCalc.rename("my_calc_module")
        assert RemoteCalc.name == "my_calc_module"
        remote_calc_renamed = RemoteCalc(owner="Runhouse", name="Runhouse_calc_new")
        assert remote_calc_renamed.summer(4, 6) == 10
        assert remote_calc_renamed.divider(35, 7) == 5
        assert remote_calc.summer(4, 6) == remote_calc_renamed.summer(4, 6)
        assert remote_calc.divider(35, 7) == remote_calc_renamed.divider(35, 7)

    @pytest.mark.usefixtures("cluster")
    @pytest.mark.level("local")
    def test_local_remote_properties(self, cluster):
        RemoteCalc = rh.module(cls=Calculator).to(cluster)
        remote_calc = RemoteCalc(owner="Runhouse", name="Runhouse_calc")
        assert RemoteCalc.name == "Calculator"
        assert remote_calc.name == "Runhouse_calc"
        assert remote_calc.remote.model == "Casio"
        assert remote_calc.remote._release_year == 2023
        assert remote_calc.local.importer == "Calculators Inc"

    @pytest.mark.level("local")
    @pytest.mark.asyncio
    async def test_fetch_class_and_properties(self, cluster):
        RemoteCalc = rh.module(cls=Calculator).to(cluster)
        owner = "Runhouse"

        # fetch of a general class (not subclass)
        remote_calc = RemoteCalc(owner=owner)
        model = remote_calc.fetch("model")
        release_year = remote_calc.fetch("_release_year")
        remote_owner = remote_calc.fetch("owner")
        model_async = await remote_calc.fetch_async("model")
        release_year_async = await remote_calc.fetch_async("_release_year")
        owner_async = await remote_calc.fetch_async("owner")
        assert model == "Casio"
        assert model_async == "Casio"
        assert release_year == 2023
        assert release_year == release_year_async
        assert remote_owner == owner_async

        # TODO: This line used to end in .fetch() . We need to update when we fix the .resolve()/serialization behavior
        remote_df_sync = SlowPandas(size=4).to(cluster)
        assert remote_df_sync.size == 4

        if cluster.on_this_cluster():
            cpu_count = os.cpu_count()
        else:
            cpu_count = int(
                cluster.run_python(["import os; print(os.cpu_count())"])[0][1]
            )
        assert remote_df_sync.cpu_count() == cpu_count

        assert await SlowPandas(size=4).to(system=cluster).fetch_async("size") == 4

    @pytest.mark.level("local")
    def test_get_or_to(self, cluster):
        RemoteCalcNew = rh.module(Calculator).get_or_to(cluster, name="new_remote_cals")
        remote_calc_new = RemoteCalcNew(owner="Runhouse_admin", name="Admin_calc")
        assert remote_calc_new.mult(5, 9) == 45
        assert remote_calc_new.remote.owner == "Runhouse_admin"
        assert remote_calc_new.name == "Admin_calc"
        RemoteCalcExisting = rh.module(Calculator).get_or_to(
            cluster, name="new_remote_cals"
        )
        remote_calc_existing = RemoteCalcExisting(
            owner="Runhouse_users", name="Users_calc"
        )
        assert remote_calc_existing.mult(6, 2) == 12
        assert remote_calc_existing.remote.owner == "Runhouse_users"
        assert remote_calc_existing.remote.name == "Users_calc"

    @pytest.mark.level("local")
    def test_refresh(self, cluster):
        RemoteCalc = rh.module(cls=Calculator).to(cluster)
        # Note: by reusing the name, we're overwriting the class module in the cluster's object store with
        # the new instance.
        remote_calc = RemoteCalc(owner="Runhouse", name=RemoteCalc.name)
        assert remote_calc.sub(33, 5) == 28
        assert remote_calc.mult(4, 9) == 36
        # Because we overwrote the class module in the cluster's object store, when we refresh it we should get
        # the instance back.
        remote_calc_refreshed = RemoteCalc.refresh()
        assert remote_calc_refreshed.system.name == cluster.name
        assert remote_calc_refreshed.remote.owner == "Runhouse"
        assert remote_calc_refreshed.mult(4, 9) == 36
        assert remote_calc_refreshed.sub(33, 5) == 28

    def get_class_name(self, cls):
        return cls._name

    def calc_square(self, a: int):
        return a * a

    @pytest.mark.level("local")
    @pytest.mark.skip(
        "TODO: Fix module serialization / resolve logic due to breakage "
        "when server is running a different Python version"
    )
    def test_resolve(self, cluster):
        remote_calc = rh.module(Calculator).to(cluster)

        # resolve method
        assert self.get_class_name(remote_calc.resolve()) == "Calculator"
        assert self.calc_square(remote_calc.resolve().mult(2, 3)) == 36
        assert self.calc_square(remote_calc.resolve().summer(2, 3)) == 25
        assert self.calc_square(remote_calc.resolve().divider(16, 2)) == 64
        assert self.calc_square(remote_calc.resolve().sub(1, 1)) == 0

        # resolved_state() method
        assert self.get_class_name(remote_calc.resolved_state()) == "Calculator"
        assert self.calc_square(remote_calc.resolved_state().mult(2, 3)) == 36
        assert self.calc_square(remote_calc.resolved_state().summer(2, 3)) == 25
        assert self.calc_square(remote_calc.resolved_state().divider(16, 2)) == 64
        assert self.calc_square(remote_calc.resolved_state().sub(1, 1)) == 0

    @pytest.mark.level("local")
    def test_save(self, cluster):
        # TODO: ask Josh for advice how to share it with a new user each time.
        if cluster.on_this_cluster():
            pytest.skip("Skipping sharing test on local cluster")

        users = ["josh@run.house"]
        remote_calc = rh.module(Calculator).to(cluster).save(name="rh_remote_calc")
        added_users, new_users, _ = remote_calc.share(
            users=users, notify_users=False, access_level="write"
        )
        assert remote_calc.name == "rh_remote_calc"
        assert added_users == {users[0]: "write"} or added_users == {}
        assert new_users == {}

    @pytest.mark.level("local")
    @pytest.mark.asyncio
    async def test_set_async(self, cluster):
        RemoteCalc = rh.module(Calculator).to(cluster)
        my_remote_calc = RemoteCalc(owner="Runhouse", name="Runhouse_remote_dev")

        # test before changing the properties values.
        assert my_remote_calc.remote.owner == "Runhouse"
        assert my_remote_calc.remote.name == "Runhouse_remote_dev"
        assert my_remote_calc.remote.model == "Casio"
        assert my_remote_calc.remote._release_year == 2023

        # test after changing the properties values.
        await my_remote_calc.set_async("owner", "Runhouse_eng")
        assert my_remote_calc.remote.owner == "Runhouse_eng"
        await my_remote_calc.set_async("_release_year", 2020)
        assert my_remote_calc.remote._release_year == 2020

        # test that the unchanged properties remained the same.
        assert my_remote_calc.remote.model == "Casio"
        assert my_remote_calc.remote.name == "Runhouse_remote_dev"

    @pytest.mark.level("unit")
    def test_signature(self):

        SlowNumpy = rh.module(SlowNumpyArray)
        assert set(SlowNumpy.signature()) == {
            "slow_iter",
            "home",
            "cpu_count",
            "print_and_log",
            "size_minus_cpus",
            "factory_constructor",
        }
        # Check that rich signature is json-able and has the same number of keys as the regular signature
        assert len(json.loads(json.dumps(SlowNumpy.signature(rich=True)))) == 6

        arr = SlowNumpy(size=5)
        assert set(arr.signature()) == {
            "slow_iter",
            "home",
            "cpu_count",
            "print_and_log",
            "size_minus_cpus",
            "factory_constructor",
        }

        df = SlowPandas(size=10)
        assert set(df.signature()) == {
            "slow_iter",
            "slow_iter_async",
            "home",
            "cpu_count",
            "cpu_count_async",
            "size_minus_cpus",
        }
        assert len(json.loads(json.dumps(df.signature(rich=True)))) == 6

        RemoteCalc = rh.module(Calculator)
        assert set(RemoteCalc.signature()) == {
            "summer",
            "asummer",
            "sub",
            "divider",
            "mult",
        }

    # SSH Clusters we can't share read only modules, since we'd have to give them an SSH key
    # in order to access the cluster. TLS clusters we can't connect to the Docker instances
    # easily.
    @pytest.mark.level("local")
    @pytest.mark.skip("Doesn't work with local clusters.")
    @pytest.mark.asyncio
    async def test_share_module_read_only(self, cluster):
        from tests.utils import friend_account

        remote_calc = rh.module(Calculator)().to(cluster, name="remote_calc_instance")

        remote_calc.share(
            users=["info@run.house"],
            access_level="read",
            notify_users=False,
        )

        with friend_account():
            await load_and_use_calculator_module(mod_name=remote_calc.rns_address)

    @pytest.mark.level("release")
    @pytest.mark.asyncio
    async def test_share_module_read_only_inter_cluster(
        self,
        ondemand_aws_https_cluster_with_auth,
        friend_account_logged_in_docker_cluster_pk_ssh,
    ):

        # Make sure cluster is not shared
        ondemand_aws_https_cluster_with_auth.revoke(users=["info@run.house"])

        # Note that this module is already constructed before it is put on the cluster
        remote_calc = (
            rh.module(Calculator)()
            .to(ondemand_aws_https_cluster_with_auth, name="remote_calc_instance")
            .save()
        )

        test_fn_remote = rh.function(load_and_use_calculator_module).to(
            friend_account_logged_in_docker_cluster_pk_ssh
        )

        remote_calc.revoke(users=["info@run.house"])

        # Expect this to raise a ValueError
        with pytest.raises(ValueError):
            await test_fn_remote(remote_calc.rns_address)

        remote_calc.share(
            users=["info@run.house"],
            access_level="read",
            notify_users=False,
        )

        await test_fn_remote(remote_calc.rns_address)

        remote_calc.revoke(users=["info@run.house"])

        # Expect this to raise a ValueError
        with pytest.raises(ValueError):
            await test_fn_remote(remote_calc.rns_address)

    @pytest.mark.level("release")
    def test_share_module_with_generator_read_only_inter_cluster(
        self,
        ondemand_aws_https_cluster_with_auth,
        friend_account_logged_in_docker_cluster_pk_ssh,
    ):
        # from tests.utils import friend_account

        size = 3
        remote_df = SlowPandas(size=size).to(
            ondemand_aws_https_cluster_with_auth, name="remote_df_ondemand_cluster"
        )

        remote_df.share(
            users=["info@run.house"],
            access_level="read",
            notify_users=False,
        )

        cpu_count = int(
            ondemand_aws_https_cluster_with_auth.run_python(
                ["import os; print(os.cpu_count())"]
            )[0][1]
        )
        test_fn = rh.fn(load_and_use_readonly_module).to(
            friend_account_logged_in_docker_cluster_pk_ssh
        )
        test_fn(mod_name=remote_df.rns_address, cpu_count=cpu_count, size=size)

    @pytest.mark.level("local")
    def test_construct_shared_module(self, cluster):
        if TEST_ORG in cluster.rns_address:
            pytest.skip(
                "Skipping test if the cluster is in the test org, as we can't share outside of the org."
            )

        from tests.utils import friend_account

        remote_calc_unconstructed = (
            rh.module(Calculator)
            .to(cluster, name="remote_calculator_unconstructed")
            .save()
        )

        cluster.revoke(
            users=["info@run.house"],
        )

        remote_calc_unconstructed.share(
            users=["info@run.house"],
            access_level="read",
            notify_users=False,
        )

        with pytest.raises(ValueError):
            with friend_account():
                construct_and_use_calculator_module(
                    mod_name=remote_calc_unconstructed.rns_address
                )

        cluster.share(
            users=["info@run.house"],
            access_level="read",
            notify_users=False,
        )

        with friend_account():
            construct_and_use_calculator_module(
                mod_name=remote_calc_unconstructed.rns_address
            )

    @pytest.mark.skip(
        "Broken in CI after we allowed higher Ray versions, probably due to Pydantic."
    )
    @pytest.mark.level("unit")
    def test_openapi_spec_generation(self):
        from openapi_core import OpenAPI

        remote_calc = rh.module(Calculator)

        # Normally we'd send the calculator to a system, save it, and then call this
        # But for testing purposes we can just pass in a name so this can run as a unit test
        spec = remote_calc.openapi_spec(spec_name="CalculatorAPI")

        # This automatically validates the spec
        openapi = OpenAPI.from_dict(spec)
        assert openapi

        # Assert that the spec can be converted to json
        assert json.loads(json.dumps(spec))

    @pytest.mark.level("local")
    def test_dependency_exception(self, cluster):
        from .exception_module import ExceptionModule

        with pytest.raises(ModuleNotFoundError) as err:
            exc_module = ExceptionModule()
            exc_module.to(cluster)
        assert "plotly" in str(err.value)

    @pytest.mark.level("local")
    @pytest.mark.asyncio
    async def test_run_async_permutations(self, cluster):
        lots_of_async_remote = rh.module(LotsOfAsync).to(cluster)
        calculator_remote = rh.module(Calculator).to(cluster)

        lots_of_async_remote_instance = lots_of_async_remote()
        calculator_remote_instance = calculator_remote()

        # Test things work natively
        assert await lots_of_async_remote_instance.asum(2, 3) == 5
        assert calculator_remote_instance.summer(2, 3) == 5

        # Test that can call with run_async set
        assert await lots_of_async_remote_instance.asum(2, 3, run_async=True) == 5
        assert lots_of_async_remote_instance.asum(2, 3, run_async=False) == 5
        assert await calculator_remote_instance.summer(2, 3, run_async=True) == 5
        assert calculator_remote_instance.summer(2, 3, run_async=False) == 5

    @pytest.mark.level("local")
    @pytest.mark.asyncio
    async def test_returns_coroutine(self, cluster):
        lots_of_async_remote = rh.module(LotsOfAsync).to(cluster)
        lots_of_async_remote_instance = lots_of_async_remote()

        # Test that can call with run_async set
        future_module = lots_of_async_remote_instance.returns_coroutine(2, 3)
        assert future_module.__class__.__name__ == "FutureModule"
        assert await future_module == 5

        # Test that can call with run_async set to True
        future_module = await lots_of_async_remote_instance.returns_coroutine(
            2, 3, run_async=True
        )
        assert future_module.__class__.__name__ == "FutureModule"
        assert await future_module == 5

        future_module = await lots_of_async_remote_instance.async_returns_coroutine(
            2, 3
        )
        assert future_module.__class__.__name__ == "FutureModule"
        assert await future_module == 5

        # Test that can call with run_async set to False
        future_module = lots_of_async_remote_instance.async_returns_coroutine(
            2, 3, run_async=False
        )
        assert future_module.__class__.__name__ == "FutureModule"
        assert await future_module == 5

    @pytest.mark.level("local")
    def test_construct_module_on_cluster(self, cluster):
        process = cluster.ensure_process_created("test_env")
        remote_constructor_module = rh.module(ConstructorModule)().to(
            cluster, process=process
        )
        remote_constructor_module.construct_module_on_cluster()

    @pytest.mark.level("local")
    def test_import_editable_package(self, cluster, installed_editable_package):

        importlib_reload(site)

        # Test that the editable package can be imported and used
        from test_fake_package import editable_package_function

        assert editable_package_function() == "Hello from the editable package!"

        # Now send this to the remote cluster and test that it can still be imported and used
        remote_editable_package_fn = rh.function(editable_package_function).to(cluster)
        assert remote_editable_package_fn() == "Hello from the editable package!"

        cluster.delete("editable_package_function")
        cluster.run_bash("pip uninstall -y test_fake_package")

    @pytest.mark.level("local")
    def test_import_editable_package_from_new_process(
        self, cluster, installed_editable_package_copy
    ):
        importlib_reload(site)

        # Test that the editable package can be imported and used
        from test_fake_package_copy import TestModuleFromPackage

        assert (
            TestModuleFromPackage.hello_world()
            == "Hello from the editable package module!"
        )

        # Now send this to the remote cluster and test that it can still be imported and used
        process = cluster.ensure_process_created("fresh_process")
        remote_editable_package_module = rh.module(TestModuleFromPackage).to(
            cluster, process=process
        )
        assert (
            remote_editable_package_module.hello_world()
            == "Hello from the editable package module!"
        )

    @pytest.mark.level("local")
    def test_module_constructed_on_cluster_is_in_same_process(self, cluster):
        process = cluster.ensure_process_created("special_process")
        remote_module = rh.module(ModuleConstructingOtherModule).to(
            system=cluster, process=process
        )
        assert remote_module.construct_and_get_process() == process

    @pytest.mark.level("local")
    def test_logs_stream_in_nested_call(self, cluster):
        size = 3
        RemoteClass = rh.module(SlowNumpyArray).to(cluster)
        remote_instance = RemoteClass(size=size, name="remote_instance1")

        # TODO test in same process (works as of 4-Nov-24)
        # remote_helper_call = rh.function(nested_call_logs_stream_helper).to(cluster)

        # Send to different process

        helper_process = cluster.ensure_process_created("helper_process")
        remote_helper_call = rh.function(nested_call_logs_stream_helper).to(
            cluster, process=helper_process
        )

        # TODO test with slow_iter call because not working with generator as of 4-Nov-24
        with capture_stdout() as stdout:
            results = remote_helper_call(remote_instance)
            out = str(stdout)
        assert len(results) == 3

        # Check that stdout of the internal module calls was captured. Skip the last result because sometimes we
        # don't catch it and it makes the test flaky.
        for i in range(size):
            assert f"Hello from the cluster stdout! {i}" in out
            assert f"Hello from the cluster logs! {i}" in out



import pytest

import runhouse as rh

from tests.conftest import init_args


@pytest.fixture(scope="session")
def package(request):
    """Parametrize over multiple packages - useful for running the same test on multiple storage types."""
    return request.getfixturevalue(request.param)


@pytest.fixture
def local_package(local_folder):
    args = {"path": local_folder.path, "install_method": "local"}
    p = rh.package(**args)
    init_args[id(p)] = args
    return p


@pytest.fixture
def s3_package(s3_folder):
    return rh.package(
        path=s3_folder.path, system=s3_folder.system, install_method="local"
    )



import subprocess
from typing import Optional

from runhouse.logger import get_logger

logger = get_logger(__name__)


def check_for_existing_ray_instance():
    ray_status_check = subprocess.run(
        ["ray", "status"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    return ray_status_check.returncode == 0


def list_actor_states(
    actor_name: Optional[str] = None,
    actor_class_name: Optional[str] = None,
    namespace: Optional[str] = "runhouse",
    state: Optional[str] = "ALIVE",
):
    import ray
    from ray.experimental.state.api import list_actors

    def filter_by(actor: "ActorState"):
        if actor_name and actor["name"] != actor_name:
            return False

        if actor_class_name and actor["class_name"] != actor_class_name:
            return False

        if namespace and actor["ray_namespace"] != namespace:
            return False

        if state and actor["state"] != state:
            return False

        return True

    return list(filter(filter_by, list_actors())) if ray.is_initialized() else []


def kill_actors(
    actor_name: Optional[str] = None,
    actor_class_name: Optional[str] = None,
    namespace: Optional[str] = None,
    gracefully: bool = True,
):
    import ray

    cluster_servlet_actor = None
    for actor in list_actor_states(actor_name, actor_class_name, namespace):
        actor_handle_to_kill = ray.get_actor(actor["name"])
        if actor["name"] == "cluster_servlet":
            cluster_servlet_actor = actor_handle_to_kill
            continue
        logger.info(f"Killing actor {actor['name']}")
        if gracefully:
            actor_handle_to_kill.__ray_terminate__.remote()
        else:
            ray.kill(actor_handle_to_kill)

    # Make sure to kill cluster_servlet last
    if cluster_servlet_actor:
        logger.info("Killing actor cluster_servlet")
        if gracefully:
            cluster_servlet_actor.__ray_terminate__.remote()
        else:
            ray.kill(cluster_servlet_actor)



from typing import Dict, List, Optional, Union

from runhouse.globals import rns_client

from runhouse.logger import get_logger
from runhouse.resources.hardware.cluster import Cluster
from runhouse.resources.hardware.constants import (
    KUBERNETES_CLUSTER_ARGS,
    ONDEMAND_COMPUTE_ARGS,
    RH_SERVER_ARGS,
    STATIC_CLUSTER_ARGS,
)
from runhouse.resources.hardware.on_demand_cluster import OnDemandCluster
from runhouse.resources.hardware.utils import (
    _config_and_args_mismatches,
    LauncherType,
    ServerConnectionType,
    setup_kubernetes,
)
from runhouse.resources.images.image import Image

logger = get_logger(__name__)


def cluster(
    name: str,
    host: Union[str, List[str]] = None,
    ssh_creds: Union[Dict, str] = None,
    ssh_port: Optional[int] = None,
    client_port: Optional[int] = None,
    server_port: Optional[int] = None,
    server_host: Optional[str] = None,
    server_connection_type: Union[ServerConnectionType, str] = None,
    ssl_keyfile: Optional[str] = None,
    ssl_certfile: Optional[str] = None,
    domain: Optional[str] = None,
    image: Optional[Image] = None,
    den_auth: bool = None,
    load_from_den: bool = True,
    dryrun: bool = False,
    **kwargs,
):
    """
    Builds an instance of :class:`Cluster`.

    * If Cluster with same name is found in Den and ``load_from_den`` is ``True``, load it down from Den
    * If arguments corresponding to ondemand clusters are provided, arguments are fed through to
      ``rh.ondemand_cluster`` factory function
    * If arguments are mismatched with loaded Cluster, return a new Cluster with the provided args

    Args:
        name (str): Name for the cluster.
        host (str or List[str], optional): Hostname (e.g. domain or name in .ssh/config), IP address, or list of IP
            addresses for the cluster (the first of which is the head node). (Default: ``None``).
        ssh_creds (Dict or str, optional): SSH credentials, passed as dictionary or the name of an ``SSHSecret`` object.
            Example: ``ssh_creds={'ssh_user': '...', 'ssh_private_key':'<path_to_key>'}`` (Default: ``None``).
        ssh_port (int, optional): Port to use for ssh. If not provided, will default to ``22``.
        client_port (int, optional): Port to use for the client. If not provided, will default to the server port.
        server_port (bool, optional): Port to use for the server. If not provided will use 80 for a
            ``server_connection_type`` or ``none``, 443 for ``tls`` and ``32300`` for all other SSH connection types.
        server_host (bool, optional): Host from which the server listens for traffic (i.e. the --host argument
            `runhouse server start` run on the cluster). Defaults to ``"0.0.0.0"`` unless connecting to the server
            with an SSH connection, in which case ``localhost`` is used. (Default: ``None``).
        server_connection_type (ServerConnectionType or str, optional): Type of connection to use for the Runhouse
            API server. ``ssh`` will use start with server via an SSH tunnel. ``tls`` will start the server
            with HTTPS on port 443 using TLS certs without an SSH tunnel. ``none`` will start the server with HTTP
            without an SSH tunnel. (Default: ``None``).
        ssl_keyfile(str, optional): Path to SSL key file to use for launching the API server with HTTPS.
            (Default: ``None``).
        ssl_certfile(str, optional): Path to SSL certificate file to use for launching the API server with HTTPS.
            (Default: ``None``).
        domain(str, optional): Domain name for the cluster. Relevant if enabling HTTPs on the cluster. (Default: ``None``).
        image (Image, optional): Default image containing setup steps to run during cluster setup. See :class:`Image`.
            (Default: ``None``)
        den_auth (bool, optional): Whether to use Den authorization on the server. If ``True``, will validate incoming
            requests with a Runhouse token provided in the auth headers of the request with the format:
            ``{"Authorization": "Bearer <token>"}``. (Default: ``None``).
        load_from_den (bool): Whether to try loading the Cluster resource from Den. (Default: ``True``)
        dryrun (bool): Whether to create the Cluster if it doesn't exist, or load a Cluster object as a dryrun.
            (Default: ``False``)

    Returns:
        Union[Cluster, OnDemandCluster]: The resulting cluster.

    Example:
        >>> # using private key
        >>> gpu = rh.cluster(host='<hostname>',
        >>>                  ssh_creds={'ssh_user': '...', 'ssh_private_key':'<path_to_key>'},
        >>>                  name='rh-a10x').save()

        >>> # using password
        >>> gpu = rh.cluster(host='<hostname>',
        >>>                  ssh_creds={'ssh_user': '...', 'password':'*****'},
        >>>                  name='rh-a10x').save()

        >>> # using the name of an SSHSecret object
        >>> gpu = rh.cluster(host='<hostname>',
        >>>                  ssh_creds="my_ssh_secret",
        >>>                  name='rh-a10x').save()

        >>> # Load cluster from above
        >>> reloaded_cluster = rh.cluster(name="rh-a10x")
    """
    if host and kwargs.get("ips"):
        raise ValueError(
            "Cluster factory method can only accept one of `host` or `ips` as an argument."
        )
    ips = kwargs.get("ips") or ([host] if isinstance(host, str) else host)

    cluster_args = locals().copy()
    cluster_args["ips"] = ips
    cluster_args["creds"] = cluster_args.pop("ssh_creds")
    cluster_args.pop("kwargs")
    cluster_args.pop("host")
    cluster_args = {k: v for k, v in cluster_args.items() if v is not None}

    # check for invalid args
    valid_kwargs = {*ONDEMAND_COMPUTE_ARGS, *KUBERNETES_CLUSTER_ARGS, *RH_SERVER_ARGS}
    unsupported_kwargs = kwargs.keys() - valid_kwargs
    if unsupported_kwargs:
        raise ValueError(
            f"Received unsupported kwargs {unsupported_kwargs}. "
            "Please refer to `rh.cluster` or `rh.ondemand_cluster` for valid input args."
        )

    if kwargs.keys() & {*ONDEMAND_COMPUTE_ARGS, *KUBERNETES_CLUSTER_ARGS}:
        if kwargs.keys() & STATIC_CLUSTER_ARGS:
            raise ValueError(
                "Received incompatible args specific to both a static and ondemand cluster."
            )
        return ondemand_cluster(
            **cluster_args,
            **kwargs,
        )

    try:
        new_cluster = Cluster.from_name(
            name, load_from_den=load_from_den, dryrun=dryrun
        )
        if isinstance(new_cluster, OnDemandCluster):
            # load from name, none of the other arguments were provided
            cluster_type = "ondemand"
        else:
            cluster_type = "static"
    except ValueError:
        new_cluster = None
        cluster_type = "unsaved"

    if cluster_type == "unsaved" and cluster_args.keys() == {
        "name",
        "load_from_den",
        "dryrun",
    }:
        raise ValueError(
            f"Cluster {name} not found in Den. Must provide cluster arguments to construct "
            "a new cluster object."
        )

    cluster_args["creds"] = cluster_args.get("creds", rns_client.default_ssh_key)
    if cluster_type == "unsaved":
        new_cluster = Cluster(**cluster_args)
    elif cluster_type == "static":
        mismatches = _config_and_args_mismatches(new_cluster.config(), cluster_args)
        server_mismatches = mismatches.keys() & RH_SERVER_ARGS
        if "ips" in mismatches:
            new_cluster = Cluster(**cluster_args)
        elif server_mismatches:
            if new_cluster.is_up():
                logger.warning(
                    "Runhouse server setting has been updated. Please run `cluster.restart_server()` "
                    f"to apply new server settings for {server_mismatches}"
                )
            new_cluster = Cluster(**cluster_args)

    new_cluster._set_connection_defaults()

    if den_auth:
        new_cluster.save()
    return new_cluster


def ondemand_cluster(
    name,
    # sky arguments
    instance_type: Optional[str] = None,
    num_nodes: Optional[int] = None,
    provider: Optional[str] = None,
    autostop_mins: Optional[int] = None,
    use_spot: bool = False,
    region: Optional[str] = None,
    memory: Union[int, str, None] = None,
    disk_size: Union[int, str, None] = None,
    num_cpus: Union[int, str, None] = None,
    accelerators: Union[int, str, None] = None,
    gpus: Union[int, str, None] = None,
    open_ports: Union[int, str, List[int], None] = None,
    vpc_name: Optional[str] = None,
    sky_kwargs: Dict = None,
    # kubernetes related arguments
    kube_namespace: Optional[str] = None,
    kube_config_path: Optional[str] = None,
    kube_context: Optional[str] = None,
    # runhouse server arguments
    server_port: int = None,
    server_host: str = None,
    server_connection_type: Union[ServerConnectionType, str] = None,
    ssl_keyfile: str = None,
    ssl_certfile: str = None,
    domain: str = None,
    image: Image = None,
    # misc arguments
    launcher: Union[LauncherType, str] = None,
    den_auth: bool = None,
    load_from_den: bool = True,
    dryrun: bool = False,
):
    """
    Builds an instance of :class:`OnDemandCluster`.

    * If Cluster with same name is found in Den and ``load_from_den`` is ``True``, load it down from Den
    * If launch arguments are mismatched with loaded Cluster, return a new Cluster with the provided args.
      These args are passed through to SkyPilot's `Resource constructor
      <https://skypilot.readthedocs.io/en/latest/reference/api.html#resources>`__: ``instance_type``,
      ``num_nodes``, ``provider``, ``use_spot``, ``region``, ``memory``, ``disk_size``, ``num_cpus``,
      ``gpus`` (``accelerators``), ``open_ports``, ``autostop_mins``, ``sky_kwargs``.
    * If runhouse related arguments are mismatched with loaded Cluster, override those Cluster properties


    Args:
        name (str): Name for the cluster, to re-use later on.
        instance_type (int, optional): Type of cloud VM type to use for the cluster, e.g. "r5d.xlarge".
            Optional, as may instead choose to specify resource requirements (e.g. memory, disk_size,
            num_cpus, gpus).
        num_nodes (int, optional): Number of nodes to use for the cluster.
        provider (str, optional): Cloud provider to use for the cluster.
        autostop_mins (int, optional): Number of minutes to keep the cluster up after inactivity,
            or ``-1`` to keep cluster up indefinitely. (Default: ``60``).
        use_spot (bool, optional): Whether or not to use spot instance. (Default: ``False``)
        region (str, optional): The region to use for the cluster. (Default: ``None``)
        memory (int or str, optional): Amount of memory to use for the cluster, e.g. `16` or "16+".
            (Default: ``None``)
        disk_size (int or str, optional): Amount of disk space to use for the cluster, e.g. `100` or "100+".
            (Default: ``None``)
        num_cpus (int or str, optional): Number of CPUs to use for the cluster, e.g. `4` or "4+". (Default: ``None``)
        gpus (int or str, optional): Type and number of GPU to use for the cluster e.g. "A101" or "L4:8".
            (Default: ``None``)
        open_ports (int or str or List[int], optional): Ports to open in the cluster's security group. Note
            that you are responsible for ensuring that the applications listening on these ports are secure.
            (Default: ``None``)
        vpc_name (str, optional): Specific VPC used for launching the cluster. If not specified,
            cluster will be launched in the default VPC.
        sky_kwargs (dict, optional): Additional keyword arguments to pass to the SkyPilot `Resource` or `launch`
            APIs. Should be a dict of the form `{"resources": {<resources_kwargs>}, "launch": {<launch_kwargs>}}`,
            where resources_kwargs and launch_kwargs will be passed to the SkyPilot Resources API (See
            `SkyPilot docs <https://skypilot.readthedocs.io/en/latest/reference/api.html#resources>`__) and `launch`
            API (See `SkyPilot docs <https://skypilot.readthedocs.io/en/latest/reference/api.html#sky-launch>`__),
            respectively. Duplicating arguments passed to the `ondemand_cluster` factory method will raise an error.
            (Default: ``None``)
        kube_namespace (str, optional): Namespace for kubernetes cluster, if applicable. (Default: ``None``)
        kube_config_path (str, optional): Path to the kube_config, for a kubernetes cluster. (Default: ``None``)
        kube_context (str, optional): Context for kubernetes cluster, if applicable. (Default: ``None``)
        server_port (bool, optional): Port to use for the server. If not provided will use 80 for a
            ``server_connection_type`` of ``none``, 443 for ``tls`` and ``32300`` for all other SSH connection types.
            (Default: ``None``)
        server_host (bool, optional): Host from which the server listens for traffic (i.e. the --host argument
            `runhouse server start` run on the cluster). Defaults to "0.0.0.0" unless connecting to the server with an SSH
            connection, in which case ``localhost`` is used. (Default: ``None``)
        server_connection_type (ServerConnectionType or str, optional): Type of connection to use for the Runhouse
            API server. ``ssh`` will use start with server via an SSH tunnel. ``tls`` will start the server
            with HTTPS on port 443 using TLS certs without an SSH tunnel. ``none`` will start the server with HTTP
            without an SSH tunnel. (Default: ``None``)
        launcher (LauncherType or str, optional): Method for launching the cluster. If set to `local`, will launch
            locally via Sky. If set to `den`, launching will be handled by Runhouse. If not provided, will be set
            to your configured default launcher, which defaults to ``local``. (Default: ``None``)
        ssl_keyfile(str, optional): Path to SSL key file to use for launching the API server with HTTPS. (Default:
            ``None``)
        ssl_certfile (str, optional): Path to SSL certificate file to use for launching the API server with HTTPS.
            (Default: ``None``)
        domain (str, optional): Domain name for the cluster. Relevant if enabling HTTPs on the cluster.
            (Default: ``None``)
        image (Image, optional): Default image containing setup steps to run during cluster setup. See :class:`Image`.
            (Default: ``None``)
        den_auth (bool, optional): Whether to use Den authorization on the server. If ``True``, will validate incoming
            requests with a Runhouse token provided in the auth headers of the request with the format:
            ``{"Authorization": "Bearer <token>"}``. (Default: ``None``).
        load_from_den (bool): Whether to try loading the Cluster resource from Den. (Default: ``True``)
        dryrun (bool): Whether to create the Cluster if it doesn't exist, or load a Cluster object as a dryrun.
            (Default: ``False``)

    Returns:
        OnDemandCluster: The resulting cluster.

    Example:
        >>> # On-Demand SkyPilot Cluster (OnDemandCluster)
        >>> gpu = rh.ondemand_cluster(name='rh-4-a100s',
        >>>                  instance_type='A100:4',
        >>>                  provider='gcp',
        >>>                  autostop_mins=-1,
        >>>                  use_spot=True,
        >>>                  region='us-east-1',
        >>>                  ).save()

        >>> # Load cluster from above
        >>> reloaded_cluster = rh.ondemand_cluster(name="rh-4-a100s")
    """
    if vpc_name and launcher == "local":
        raise ValueError(
            "Custom VPCs are not supported with local launching. To use a custom VPC, please use the "
            "Den launcher. For more information see "
            "https://www.run.house/docs/installation-setup#den-launcher"
        )

    cluster_args = locals().copy()
    cluster_args = {k: v for k, v in cluster_args.items() if v is not None}
    if "accelerators" in cluster_args:
        logger.warning(
            "``accelerators`` argument has been deprecated. Please use ``gpus`` argument instead."
        )
        cluster_args["gpus"] = cluster_args.pop("accelerators")

    try:
        new_cluster = Cluster.from_name(
            name, load_from_den=load_from_den, dryrun=dryrun
        )
        cluster_type = "ondemand"
    except ValueError:
        new_cluster = None
        cluster_type = "unsaved"

    if cluster_args.keys() & KUBERNETES_CLUSTER_ARGS:
        setup_kubernetes(**cluster_args)

    if cluster_type == "unsaved":
        if cluster_args.keys() == {"name", "use_spot", "load_from_den", "dryrun"}:
            raise ValueError(
                f"OndemandCluster {name} not found in Den. Must provide cluster arguments to construct "
                "a new cluster object."
            )
        new_cluster = OnDemandCluster(**cluster_args)
    elif cluster_type == "ondemand":
        mismatches = _config_and_args_mismatches(new_cluster.config(), cluster_args)
        compute_mismatches = {
            k: v
            for k, v in mismatches.items()
            if k in {*ONDEMAND_COMPUTE_ARGS, *KUBERNETES_CLUSTER_ARGS}
        }
        server_mismatches = {k: v for k, v in mismatches.items() if k in RH_SERVER_ARGS}
        new_autostop_mins = compute_mismatches.pop("autostop_mins", None)

        if mismatches and new_cluster.is_up():
            # cluster is up - throw error if launch compute mismatches, but allow server / autostop min updates
            if compute_mismatches:
                raise ValueError(
                    f"Cluster {name} is up, but received argument mismatches for compute: {compute_mismatches.keys()}. "
                    "Please construct a new cluster object or ensure that the arguments match."
                )
            if new_autostop_mins:
                logger.info(f"Updating autostop mins for cluster {name}")
                new_cluster.autostop_mins = new_autostop_mins
            if server_mismatches:
                new_cluster._update_values(server_mismatches)
                logger.warning(
                    "Runhouse server setting has been updated. Please run `cluster.restart_server()` "
                    f"to apply new server settings for {server_mismatches.keys()}"
                )
        else:
            # cluster is down
            # - construct new cluster if launch compute mismatches
            # - if compute matches/empty but server or autostop mismatches, override just those values
            if compute_mismatches:
                new_cluster = OnDemandCluster(**cluster_args)
            elif server_mismatches:
                new_cluster._update_values(server_mismatches)
            if new_autostop_mins:
                new_cluster._autostop_mins = new_autostop_mins

    new_cluster._set_connection_defaults()
    if den_auth:
        new_cluster.save()
    return new_cluster



import time
from typing import List, Optional

from runhouse.resources.distributed.supervisor import Supervisor

from runhouse.resources.module import Module


class DistributedPool(Supervisor):
    def __init__(
        self, name, replicas: List[Module] = None, max_concurrency: int = 1, **kwargs
    ):
        super().__init__(name=name, **kwargs)
        self._replicas = replicas or []
        self._max_concurrency = max_concurrency
        self._available_replicas = list(
            range(len(self._replicas) * self._max_concurrency)
        )

    def _compute_signature(self, rich=False):
        return self.local._replicas[0].signature(rich=rich)

    def forward(self, item, timeout: Optional[int] = None, *args, **kwargs):
        time_waited = 0
        while not self._available_replicas:
            if timeout == 0:
                raise TimeoutError("No available replicas.")
            if timeout is not None and time_waited >= timeout:
                raise TimeoutError("Timed out waiting for a replica to be available.")
            time.sleep(0.25)
            time_waited += 0.25
        worker_idx = self._available_replicas.pop(0)
        worker = self._replicas[worker_idx // self._max_concurrency]
        method = getattr(worker, item)
        res = method(*args, **kwargs)
        self._available_replicas.append(worker_idx)
        return res

    def __call__(self, *args, **kwargs):
        return self.call(*args, **kwargs)



import importlib
import math
import subprocess
import time

from datetime import datetime, timezone
from enum import Enum
from typing import Any, Dict, List

import requests

import rich
import typer
from rich.table import Table

import runhouse as rh

from runhouse.constants import (
    BULLET_UNICODE,
    DEFAULT_PROCESS_NAME,
    DOUBLE_SPACE_UNICODE,
    HOUR,
    LAST_ACTIVE_AT_TIMEFRAME,
    MAX_CLUSTERS_DISPLAY,
    SERVER_START_CMD,
    START_NOHUP_CMD,
    START_SCREEN_CMD,
)

from runhouse.logger import get_logger

from runhouse.resources.hardware.utils import ClusterStatus, LauncherType
from runhouse.servers.obj_store import ObjStoreError

logger = get_logger(__name__)


####################################################################################################
# Cluster list utils
####################################################################################################
class ClusterStatusColors(str, Enum):
    RUNNING = "[green]Running[/green]"
    INITIALIZING = "[yellow]Initializing[/yellow]"
    TERMINATED = "[red]Terminated[/red]"
    UNKNOWN = "Unknown"

    @classmethod
    def get_status_color(cls, status: str):
        try:
            return getattr(cls, status.upper()).value
        except AttributeError:
            return cls.UNKNOWN.value


class AutoStopColors:
    @classmethod
    def format_autostop(cls, autostop_value: int = None):
        if autostop_value is None:
            # Leave blank if autostop is not set or not relevant for the cluster type
            return None

        if autostop_value == -1:
            return "[red]Suspended[/red]"
        else:
            return f"[yellow]{autostop_value}[/yellow]"


def create_output_table(
    total_clusters: int,
    running_clusters: int,
    displayed_clusters: int,
    filters_requested: bool,
):
    """The cluster list is printed as a table, this method creates it."""
    from runhouse.globals import rns_client

    displayed_running_clusters = (
        running_clusters
        if running_clusters < displayed_clusters
        else displayed_clusters
    )
    table_title = (
        f"[bold cyan]Clusters for {rns_client.username} "
        f"(Running: {displayed_running_clusters}/{running_clusters}, "
        f"Total Displayed: {displayed_clusters}/{total_clusters})[/bold cyan]"
    )

    table = Table(title=table_title)

    if not filters_requested:
        table.caption = (
            f"[reset]Showing clusters that were active in the "
            f"last {int(LAST_ACTIVE_AT_TIMEFRAME / HOUR)} hours."
        )
        table.caption_justify = "left"

    if displayed_clusters == MAX_CLUSTERS_DISPLAY:
        link_to_clusters_in_den = "[reset]The full list of clusters can be viewed at https://www.run.house/resources?type=cluster"
        if table.caption:
            table.caption += f"\n{link_to_clusters_in_den}"
        else:
            table.caption = link_to_clusters_in_den
        table.caption_justify = "left"

    # Add columns to the table
    table.add_column("Name", justify="left", no_wrap=True)
    table.add_column("Cluster Type", justify="left", no_wrap=True)
    table.add_column("Status", justify="left")
    table.add_column("Last Active (UTC)", justify="left")
    table.add_column("Autostop (Mins)", justify="left")

    return table


def add_cluster_as_table_row(table: Table, rh_cluster: dict):
    """Adding an info of a single cluster to the output table."""
    table.add_row(
        rh_cluster.get("Name"),
        rh_cluster.get("Cluster Type"),
        rh_cluster.get("Status"),
        rh_cluster.get("Last Active (UTC)"),
        rh_cluster.get("Autostop (Mins)"),
    )

    return table


def add_clusters_to_output_table(table: Table, clusters: List[Dict]):
    """Adding clusters info to the output table."""
    for rh_cluster in clusters:
        last_active_at = rh_cluster.get("Last Active (UTC)")
        if not last_active_at:  # case when last_active_at == None
            last_active_at_no_offset = last_active_at
        else:
            last_active_at_no_offset = str(last_active_at).split("+")[
                0
            ]  # The split is required to remove the offset (according to UTC)
        rh_cluster["Last Active (UTC)"] = last_active_at_no_offset
        rh_cluster["Status"] = ClusterStatusColors.get_status_color(
            rh_cluster.get("Status")
        )
        autostop_value = rh_cluster.get("Autostop")
        rh_cluster["Autostop (Mins)"] = AutoStopColors.format_autostop(
            int(autostop_value) if autostop_value else None
        )
        table = add_cluster_as_table_row(table, rh_cluster)


def condense_resource_type(resource_type: str):
    """
    status helping function. transforms a str form runhouse.resources.{X.Y...}.resource_type to runhouse.resource_type
    """
    try:
        resource_type = resource_type.split(".")[-1]
        getattr(importlib.import_module("runhouse"), resource_type)
        return f"runhouse.{resource_type}"
    except AttributeError:
        return resource_type


####################################################################################################
# Cluster status utils
####################################################################################################
class StatusType(str, Enum):
    server = "server"
    cluster = "cluster"


def print_cluster_config(cluster_config: Dict, status_type: str = StatusType.cluster):
    """
    Helping function to the `_print_status` which prints the relevant info from the cluster config.
    """
    from runhouse.main import console

    top_level_config = [
        "server_port",
        "den_auth",
        "server_connection_type",
    ]

    backend_config = ["domain", "server_host"]
    if status_type == StatusType.cluster:
        backend_config = backend_config + ["resource_subtype", "ips"]

        if cluster_config.get("resource_subtype") != "Cluster":
            backend_config.append("autostop_mins")

    for key in top_level_config:
        console.print(
            f"{BULLET_UNICODE} {key.replace('_', ' ')}: {cluster_config[key]}"
        )

    if status_type == StatusType.cluster:
        console.print(f"{BULLET_UNICODE} backend config:")

    for key in backend_config:
        if key == "autostop_mins" and cluster_config[key] == -1:
            console.print(
                f"{DOUBLE_SPACE_UNICODE}{BULLET_UNICODE} {key.replace('_', ' ')}: autostop disabled"
            )
        else:
            if (
                key == "ips"
                and cluster_config.get("resource_subtype") == "OnDemandCluster"
            ):
                val = cluster_config.get("compute_properties", {}).get("ips", [])
            else:
                val = cluster_config.get(key, None)

            # don't print keys whose values are None
            if val is None:
                continue

            console.print(
                f"{DOUBLE_SPACE_UNICODE}{BULLET_UNICODE} {key.replace('_', ' ')}: {val}"
            ) if status_type == StatusType.cluster else console.print(
                f"{BULLET_UNICODE} {key.replace('_', ' ')}: {val}"
            )


def print_envs_info(servlet_processes: Dict[str, Dict[str, Any]], current_cluster):
    """
    Prints info about the envs in the current_cluster: resources in each env, the CPU usage and GPU usage of the env
    (if exists)
    """
    from runhouse.main import console

    # Print headline
    envs_in_cluster_headline = "Serving 🍦 :"
    console.print(envs_in_cluster_headline)

    env_resource_mapping = {
        env: servlet_processes[env].get("env_resource_mapping", {})
        for env in servlet_processes
    }

    if len(env_resource_mapping) == 0:
        console.print("This cluster has no environment nor resources.")

    first_envs_to_print = []

    # First: if the default env does not have resources, print it.
    default_process_name = DEFAULT_PROCESS_NAME
    if len(env_resource_mapping[default_process_name]) == 0:
        # case where the default env doesn't hve any other resources, apart from the default env itself.
        console.print(f"{BULLET_UNICODE} {default_process_name}")
        console.print(
            f"{DOUBLE_SPACE_UNICODE}This process has only python packages installed, if provided. No "
            "resources were found."
        )

    else:
        # if the default env has other resources make sure it gets printed first
        first_envs_to_print = [default_process_name]

    # Make sure to print envs with no resources first.
    # (the only resource they have is a runhouse.env, which is the env itself).
    first_envs_to_print = first_envs_to_print + [
        env_name
        for env_name in env_resource_mapping
        if (
            len(env_resource_mapping[env_name]) == 0
            and env_name != default_process_name
            and env_resource_mapping[env_name]
        )
    ]

    # Now, print the envs.
    # If the env have packages installed, that means that it contains an env resource. In that case:
    # * If the env contains only itself, we will print that the env contains only the installed packages.
    # * Else, we will print the resources (rh.function, th.module) associated with the env.
    envs_to_print = first_envs_to_print + [
        env_name
        for env_name in env_resource_mapping
        if env_name not in first_envs_to_print + [default_process_name]
    ]

    for env_name in envs_to_print:
        resources_in_env = env_resource_mapping[env_name]
        env_process_info = servlet_processes[env_name]

        pid, node_name = env_process_info.get("pid", None), env_process_info.get(
            "node_name", None
        )

        # if there is no info about the process, don't print it in the status output
        if not pid or not node_name:
            continue

        env_name_txt = f"{BULLET_UNICODE} {env_name} | pid: {pid} | node: {node_name}"
        console.print(env_name_txt)

        # Print CPU info
        env_cpu_info = env_process_info.get("env_cpu_usage")
        if env_cpu_info:

            # convert bytes to GB
            memory_usage_gb = round(
                int(env_cpu_info["used_memory"]) / (1024**3),
                2,
            )
            total_cluster_memory = math.ceil(
                int(env_cpu_info["total_memory"]) / (1024**3)
            )
            cpu_memory_usage_percent = round(
                float(env_cpu_info["used_memory"] / env_cpu_info["total_memory"]) * 100,
                2,
            )
            cpu_usage_percent = round(float(env_cpu_info["utilization_percent"]), 2)

            cpu_usage_summary = f"{DOUBLE_SPACE_UNICODE}CPU: {cpu_usage_percent}% | Memory: {memory_usage_gb} / {total_cluster_memory} Gb ({cpu_memory_usage_percent}%)"

        else:
            cpu_usage_summary = (
                f"{DOUBLE_SPACE_UNICODE}CPU: This process did not use CPU memory."
            )

        console.print(cpu_usage_summary)

        # Print GPU info
        env_gpu_info = env_process_info.get("env_gpu_usage")

        # sometimes the cluster has no GPU, therefore the env_gpu_info is an empty dictionary.
        if env_gpu_info:
            # get the gpu usage info, and convert it to GB.
            total_gpu_memory = math.ceil(
                float(env_gpu_info.get("total_memory")) / (1024**3)
            )
            used_gpu_memory = round(
                float(env_gpu_info.get("used_memory")) / (1024**3), 2
            )
            gpu_memory_usage_percent = round(
                float(used_gpu_memory / total_gpu_memory) * 100, 2
            )
            gpu_usage_summery = f"{DOUBLE_SPACE_UNICODE}GPU Memory: {used_gpu_memory} / {total_gpu_memory} Gb ({gpu_memory_usage_percent}%)"
            console.print(gpu_usage_summery)

        resources_in_env = [
            {resource: resources_in_env[resource]}
            for resource in resources_in_env
            if resource is not env_name
        ]

        if len(resources_in_env) == 0:
            # No resources were found in the env, only the associated installed python reqs were installed.
            console.print(
                f"{DOUBLE_SPACE_UNICODE}No objects are stored in this process."
            )

        else:
            for resource in resources_in_env:
                for resource_name, resource_info in resource.items():
                    resource_type = condense_resource_type(
                        resource_info.get("resource_type")
                    )

                    if resource_type == "runhouse.Env":
                        resource_type = None

                    active_function_calls = resource_info.get("active_function_calls")
                    resource_info_str = (
                        f"{DOUBLE_SPACE_UNICODE}{BULLET_UNICODE} {resource_name}"
                    )
                    if resource_type:
                        resource_info_str = resource_info_str + f" ({resource_type})"

                    if resource_type == "runhouse.Function" and active_function_calls:
                        func_start_time_utc = active_function_calls[0].get(
                            "start_time", None
                        )

                        # casting func_start_time_utc to datetime format
                        func_start_time_utc = datetime.fromtimestamp(
                            func_start_time_utc, tz=timezone.utc
                        )

                        # func_end_time_utc = current time. Making sure it is in the same format as func_start_time_utc,
                        # so we could calculate function's running time.
                        func_end_time_utc = datetime.fromtimestamp(
                            time.time(), tz=timezone.utc
                        )

                        func_running_time = (
                            func_end_time_utc - func_start_time_utc
                        ).total_seconds()

                        is_func_running: str = (
                            f" [italic bright_green]Running for {func_running_time} "
                            f"seconds[/italic bright_green]"
                        )

                    elif (
                        resource_type == "runhouse.Function"
                        and not active_function_calls
                    ):
                        is_func_running: str = " [italic bright_yellow]Currently not running[/italic bright_yellow]"

                    else:
                        is_func_running: str = ""

                    resource_info_str = resource_info_str + is_func_running

                    console.print(resource_info_str)


def print_cloud_properties(cluster_config: dict):
    from runhouse.main import console

    cloud_properties = cluster_config.get("compute_properties", None)
    if not cloud_properties:
        return

    cloud = cloud_properties.get("cloud")
    instance_type = cloud_properties.get("instance_type")
    region = cloud_properties.get("region")
    cost_per_hour = cloud_properties.get("cost_per_hour")

    is_gpu = cluster_config.get("is_gpu", False)
    cost_emoji = "💰" if is_gpu else "💸"

    num_of_cpus = cloud_properties.get("num_cpus") or len(cluster_config.get("ips"))
    num_of_gpus = 0
    cluster_gpus = cloud_properties.get("gpus", None)
    gpu_types = set()
    if cluster_gpus:
        for k, v in cluster_gpus.items():
            num_of_gpus = num_of_gpus + int(v)
            gpu_types.add(k)

    console.print(
        f"[reset]🤖 {cloud} {instance_type} cluster | 🌍 {region} | {cost_emoji} ${cost_per_hour}/hr"
    )
    cpus_gpus_info_str = f"CPUs: {int(float(num_of_cpus))}"
    if num_of_gpus > 0:
        gpu_types_str = gpu_types.pop()
        for gpu_type in gpu_types:
            gpu_types_str = gpu_types_str + f", {gpu_type}"
        cpus_gpus_info_str = (
            cpus_gpus_info_str + f" | GPUs: {num_of_gpus} (Type(s): {gpu_types_str})"
        )
    console.print(f"[reset]{cpus_gpus_info_str}")


def print_status(status_data: dict, current_cluster) -> None:
    """Prints the status of the cluster to the console"""
    from runhouse.globals import rns_client
    from runhouse.main import console

    cluster_config = status_data.get("cluster_config")
    servlet_processes = status_data.get("env_servlet_processes")

    cluster_name = cluster_config.get("name", None)
    if cluster_name:
        cluster_uri = rns_client.format_rns_address(cluster_name)
        cluster_link_in_den_ui = f"https://www.run.house/resources/{cluster_uri}"
        cluster_name_hyperlink = rich.markdown.Text(
            cluster_name, style=f"link {cluster_link_in_den_ui} white"
        )
        console.print(cluster_name_hyperlink)

    is_gpu = cluster_config.get("is_gpu", False)

    # print headline
    daemon_headline_txt = (
        "\N{smiling face with horns} Runhouse server is running \N{Runner}"
    )
    console.print(daemon_headline_txt, style="bold royal_blue1")

    console.print(f"[reset]Runhouse v{status_data.get('runhouse_version')}")
    print_cloud_properties(cluster_config)
    console.print(f"[reset]server pid: {status_data.get('server_pid')}")

    # Print relevant info from cluster config.
    print_cluster_config(cluster_config)

    # print general cpu and gpu utilization
    cluster_gpu_utilization: float = status_data.get("server_gpu_utilization")

    # Note: GPU utilization can be none if the cluster was not using its GPU when cluster.status() was invoked
    if cluster_gpu_utilization is None and is_gpu:
        cluster_gpu_utilization: float = 0.0

    cluster_cpu_utilization: float = status_data.get("server_cpu_utilization")

    server_util_info = (
        f"CPU Utilization: {round(cluster_cpu_utilization, 2)}% | GPU Utilization: {round(cluster_gpu_utilization, 2)}%"
        if is_gpu
        else f"CPU Utilization: {round(cluster_cpu_utilization, 2)}%"
    )
    console.print(server_util_info)

    # print the environments in the cluster, and the resources associated with each environment.
    print_envs_info(servlet_processes, current_cluster)


def get_local_or_remote_cluster(cluster_name: str = None, exit_on_error: bool = True):
    from runhouse.main import console

    if cluster_name:
        try:
            current_cluster = rh.cluster(name=cluster_name, dryrun=True)
        except ValueError as e:
            console.print("Cluster not found in Den.")
            if exit_on_error:
                raise typer.Exit(1)
            raise e

        if isinstance(current_cluster, rh.OnDemandCluster):

            # in case we called current_cluster.up() on a local cluster, we need to update the cluster_status,
            # because its being updated properly only if we call up_if_not()
            if (
                current_cluster.launcher == LauncherType.LOCAL
                and current_cluster.cluster_status != ClusterStatus.RUNNING
            ):
                current_cluster._fetch_sky_status_and_update_cluster_status(
                    refresh=True
                )

            if current_cluster.cluster_status == ClusterStatus.INITIALIZING:
                console.print(
                    f"[reset]{cluster_name} is being initialized. Please wait for it to finish, or run [reset][bold italic]`runhouse cluster up {cluster_name} -f`[/bold italic] to abort the initialization and relaunch."
                )
                raise typer.Exit(0)

        if not current_cluster.is_up():
            console.print(
                f"Cluster [reset]{cluster_name} is not up. If it's an on-demand cluster, you can run "
                f"[reset][bold italic]`runhouse cluster up {cluster_name}`[/bold italic] to bring it up automatically."
            )
            if exit_on_error:
                raise typer.Exit(1)
        try:
            if current_cluster._http_client:
                current_cluster._http_client.check_server()
        except requests.exceptions.ConnectionError:
            console.print(
                f"Could not connect to the server on cluster [reset]{cluster_name}. Check that the server is up with "
                f"[reset][bold italic]`runhouse cluster status {cluster_name}`[/bold italic] or"
                f" [bold italic]`sky status -r`[/bold italic] for locally launched on-demand clusters."
            )
            if exit_on_error:
                raise typer.Exit(1)
        return current_cluster

    try:
        cluster_or_local = rh.here
    except ObjStoreError:
        console.print("Could not connect to Runhouse server. Is it up?")
        if exit_on_error:
            raise typer.Exit(1)

    if cluster_or_local == "file":
        # If running outside the cluster must specify a cluster name
        console.print(
            "Please specify a `cluster_name` or run [reset][bold italic]`runhouse server start`[/bold italic] to start "
            "a Runhouse server locally."
        )
        if exit_on_error:
            raise typer.Exit(1)
    elif not cluster_or_local:
        console.print(
            "\N{smiling face with horns} Runhouse Daemon is not running... \N{No Entry} \N{Runner}. "
            "Start it with [reset][bold italic]`runhouse server restart`[/bold italic] or specify a remote "
            "cluster to poll with [reset][bold italic]`runhouse cluster status <cluster_name>`[/bold italic]."
        )
        if exit_on_error:
            raise typer.Exit(1)

    else:
        # we are inside the cluster
        current_cluster = cluster_or_local  # cluster_or_local = rh.here

    return current_cluster


####################################################################################################
# General utils
####################################################################################################
class LogsSince(str, Enum):
    # Note: All options are represented in minutes
    one = 1
    five = 5
    ten = 10
    sixty = 60  # one hour
    one_eighty = int(3 * (HOUR / 60))  # three hours
    three_sixty = int(6 * (HOUR / 60))  # six hours
    day = int(24 * (HOUR / 60))  # one day


def is_command_available(cmd: str) -> bool:
    """Checks if a command is available on the system."""
    cmd_check = subprocess.run(
        f"command -v {cmd}", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    available = cmd_check.returncode == 0
    if not available:
        logger.info(f"{cmd} is not available on the system.")
    return available


def get_wrapped_server_start_cmd(flags: List[str], screen: bool, nohup: bool):
    """Add flags to the base server start command"""
    if screen:
        wrapped_cmd = START_SCREEN_CMD
    elif nohup:
        wrapped_cmd = START_NOHUP_CMD
    else:
        wrapped_cmd = SERVER_START_CMD

    if flags:
        flags_str = "".join(flags)
        wrapped_cmd = wrapped_cmd.replace(
            SERVER_START_CMD, SERVER_START_CMD + flags_str
        )

    return wrapped_cmd


def check_ray_installation():
    try:
        import ray  # noqa

    except ImportError:
        raise ImportError(
            "Ray is required for this command. "
            'You can install Ray and other server dependencies using `pip install "runhouse[server]"`'
        )
